from __future__ import annotations

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from silx.io.url import DataUrl

import tomwer.core.process.edit.darkflatpatch
from tomwer.gui.edit.dkrfpatch import DarkRefPatchWidget

from ...orange.managedprocess import TomwerWithStackStack

_logger = logging.getLogger(__name__)


class DarkFlatPatchOW(
    TomwerWithStackStack,
    ewokstaskclass=tomwer.core.process.edit.darkflatpatch.DarkFlatPatchTask,
):
    """
    Widget to define on the fly the image_key of a NXtomoScan
    """

    name = "dark-flat-patch"
    id = "orange.widgets.tomwer.edit.DarkFlatPatchOW.DarkFlatPatchOW"
    description = "Interface to patch dark and flat to an existing NXTomo" "entry"
    icon = "icons/patch_dark_flat.svg"
    priority = 25
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "dark",
        "patch",
        "ref",
        "flat",
    ]

    want_main_area = True
    resizing_enabled = True

    _urlsSetting = Setting(dict())
    _ewoks_inputs_to_hide_from_orange = ("configuration", "serialize_output_data")

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        layout = gui.vBox(self.mainArea, self.name).layout()

        self.widget = DarkRefPatchWidget(parent=self)
        layout.addWidget(self.widget)

        self.setConfiguration(self._urlsSetting)

        # connect signal / slot
        self.task_executor_queue.sigComputationStarted.connect(self._startProcessing)
        self.task_executor_queue.sigComputationEnded.connect(self._endProcessing)
        self.widget.sigConfigurationChanged.connect(self._updateSettings)
        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError

    def reprocess(self, dataset):
        self.set_default_input(
            "data",
            dataset,
        )

    def get_task_inputs(self) -> dict:
        task_inputs = super().get_task_inputs()
        task_inputs.update(
            {
                "configuration": self.getConfiguration(),
                "serialize_output_data": False,  # avoid serializing when using orange
            }
        )
        return task_inputs

    def getConfiguration(self):
        def cast_url_to_str(url: DataUrl | None):
            if url is None:
                return None
            elif not isinstance(url, DataUrl):
                raise TypeError(
                    f"url is expected to be an optional DataUrl. {type(url)} provided instead"
                )
            else:
                return url.path()

        return {
            "darks_start": cast_url_to_str(self.widget.getStartDarkUrl()),
            "flats_start": cast_url_to_str(self.widget.getStartFlatUrl()),
            "darks_end": cast_url_to_str(self.widget.getEndDarkUrl()),
            "flats_end": cast_url_to_str(self.widget.getEndFlatUrl()),
        }

    def setConfiguration(self, config):
        if config is None:
            return
        self.widget.clear()
        url_keys = ("darks_start", "flats_start", "darks_end", "flats_end")
        url_index_keys = (
            "darks_start_index",
            "flats_start_index",
            "darks_end_index",
            "flats_end_index",
        )
        setters = (
            self.widget.setStartDarkUrl,
            self.widget.setStartFlatUrl,
            self.widget.setEndDarkUrl,
            self.widget.setEndFlatUrl,
        )
        for url_key, url_idx_key, setter in zip(url_keys, url_index_keys, setters):
            if url_key in config:
                index = config.get(url_idx_key, 0)
                url = config[url_key]
                if isinstance(url, str):
                    url = DataUrl(path=url)
                if url not in (None, ""):
                    assert isinstance(
                        url, DataUrl
                    ), "url is expected to be an url. Either as a str or as an instance of DataUrl"
                    try:
                        setter(url=url, series_index=index)
                    except Exception as e:
                        _logger.error(e)

    def _updateSettings(self):
        self._urlsSetting = self.getConfiguration()
        self._urlsSetting.update(
            {
                "darks_start_index": self.widget.getStartDarkIndex(),
                "flats_start_index": self.widget.getStartFlatIndex(),
                "darks_end_index": self.widget.getEndDarkIndex(),
                "flats_end_index": self.widget.getEndFlatIndex(),
            }
        )

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            scan = task_executor.current_task.inputs.data
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            _logger.error(f"failed to handle task finished callback. Reason is {e}")



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile

import fabio
import numpy
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefsTask
from tomwer.core.process.reconstruction.darkref.params import DKRFRP, ReduceMethod
from tomwer.core.utils.scanutils import MockEDF
from tomwer.tests.datasets import TomwerCIDatasets


class TestDarkRefsBehavior(TestCaseQt):
    """Test that the Darks and reference are correctly computed from the
    DarksRefs class
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self.datasetsID = "test10"
        self.tmpDir = tempfile.mkdtemp()
        self.thRef = {}
        """number of theoretical ref file the algorithm should create"""
        self.thDark = {}
        """number of theoretical dark file the algorithm should create"""
        self.dataset_folder = os.path.join(self.tmpDir, self.datasetsID)

        dataDir = TomwerCIDatasets.get_dataset(
            f"edf_datasets/{self.datasetsID}",
        )
        shutil.copytree(dataDir, self.dataset_folder)
        files = os.listdir(self.dataset_folder)
        for _f in files:
            if _f.startswith(("refHST", "darkHST", "dark.edf")):
                os.remove(os.path.join(self.dataset_folder, _f))

        self.recons_params = DKRFRP()
        self.recons_params._set_remove_opt(False)

    def tearDown(self):
        self.qapp.processEvents()
        self.darkRef = None
        shutil.rmtree(self.tmpDir)

    def testDarkCreation(self):
        """Test that the dark is correctly computed"""
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEDIAN

        dar_ref_process = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "force_sync": True,
                "darkhst_prefix": "darkHST",
                "data": self.dataset_folder,
                "serialize_output_data": False,
            }
        )
        dar_ref_process.run()
        self.qapp.processEvents()
        if os.path.basename(self.dataset_folder) == "test10":
            self.assertTrue("darkend0000.edf" in os.listdir(self.dataset_folder))
            self.assertTrue("dark.edf" in os.listdir(self.dataset_folder))
            self.assertEqual(
                len(
                    dar_ref_process.getDarkHSTFiles(
                        self.dataset_folder, prefix=self.recons_params.dark_prefix
                    )
                ),
                1,
            )
            self.assertEqual(
                len(
                    dar_ref_process.getRefHSTFiles(
                        self.dataset_folder, prefix=self.recons_params.flat_prefix
                    )
                ),
                0,
            )

    def testRefCreation(self):
        """Test that the dark is correctly computed"""
        self.recons_params.flat_calc_method = ReduceMethod.MEDIAN
        self.recons_params.dark_calc_method = ReduceMethod.NONE

        dar_ref_process = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "force_sync": True,
                "darkhst_prefix": "darkHST",
                "data": self.dataset_folder,
                "serialize_output_data": False,
            }
        )

        dar_ref_process.run()
        self.qapp.processEvents()
        self.assertTrue("darkend0000.edf" in os.listdir(self.dataset_folder))
        self.assertFalse("dark0000.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("refHST0000.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("refHST0020.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("ref0000_0000.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("ref0000_0020.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("ref0001_0000.edf" in os.listdir(self.dataset_folder))
        self.assertTrue("ref0001_0020.edf" in os.listdir(self.dataset_folder))


class TestRefCalculationOneSerie(TestCaseQt):
    """
    Make sure the calculation is correct for the dark and flat field
    according to the method used.
    """

    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        n_scans = 5
        n_info = 1
        n_xml = 1
        MockEDF.fastMockAcquisition(self.tmp_dir, n_radio=n_scans)
        reFiles = {}
        data1 = numpy.zeros((20, 10))
        data2 = numpy.zeros((20, 10)) + 100
        reFiles["ref0000_0000.edf"] = data1
        reFiles["ref0001_0000.edf"] = data2
        reFiles["ref0002_0000.edf"] = data2
        reFiles["ref0003_0000.edf"] = data2
        for refFile in reFiles:
            file_desc = fabio.edfimage.EdfImage(data=reFiles[refFile])
            file_desc.write(os.path.join(self.tmp_dir, refFile))
        assert len(os.listdir(self.tmp_dir)) is (
            len(reFiles) + n_scans + n_xml + n_info
        )

        self.recons_params = DKRFRP()
        self.darkRef = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "data": self.tmp_dir,
                "serialize_output_data": False,
            }
        )
        self.darkRef.setForceSync(True)
        self.recons_params.flat_pattern = "ref*.*[0-9]{3,4}_[0-9]{3,4}"

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        super().tearDown()

    def testRefMedianCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.MEDIAN
        self.recons_params.dark_calc_method = ReduceMethod.NONE
        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "refHST0000.edf")
        self.assertTrue(os.path.isfile(refHST))
        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 100)
        )

    def testRefMeanCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.MEAN
        self.recons_params.dark_calc_method = ReduceMethod.NONE
        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "refHST0000.edf")
        self.assertTrue(os.path.isfile(refHST))
        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 75)
        )


class TestRefCalculationThreeSerie(TestCaseQt):
    """
    Make sure the calculation is correct for the dark and flat field
    according to the method used.
    """

    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        MockEDF.fastMockAcquisition(folder=self.tmp_dir, n_radio=1)
        reFiles = {}
        self.seriesList = (0, 10, 200)
        for series in self.seriesList:
            data1 = numpy.zeros((20, 10)) + series
            data2 = numpy.zeros((20, 10)) + 100 + series
            reFiles["ref0000_" + str(series).zfill(4) + ".edf"] = data1
            reFiles["ref0001_" + str(series).zfill(4) + ".edf"] = data2
            reFiles["ref0002_" + str(series).zfill(4) + ".edf"] = data2
            reFiles["ref0003_" + str(series).zfill(4) + ".edf"] = data2
            for refFile in reFiles:
                file_desc = fabio.edfimage.EdfImage(data=reFiles[refFile])
                file_desc.write(os.path.join(self.tmp_dir, refFile))

        self.recons_params = DKRFRP()
        self.darkRef = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "data": self.tmp_dir,
                "force_sync": True,
                "serialize_output_data": False,
            }
        )
        self.recons_params.flat_pattern = "ref*.*[0-9]{3,4}_[0-9]{3,4}"

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        super().tearDown()

    def testRefMedianCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.MEDIAN
        self.recons_params.dark_calc_method = ReduceMethod.NONE
        self.darkRef.run()
        for serie in self.seriesList:
            refHST = os.path.join(self.tmp_dir, "refHST" + str(serie).zfill(4) + ".edf")
            self.assertTrue(os.path.isfile(refHST))
            self.assertTrue(
                numpy.array_equal(
                    fabio.open(refHST).data, numpy.zeros((20, 10)) + 100 + serie
                )
            )

    def testRefMeanCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.MEAN
        self.recons_params.dark_calc_method = ReduceMethod.NONE
        self.darkRef.run()
        for series in self.seriesList:
            refHST = os.path.join(
                self.tmp_dir, "refHST" + str(series).zfill(4) + ".edf"
            )
            self.assertTrue(os.path.isfile(refHST))
            self.assertTrue(
                numpy.array_equal(
                    fabio.open(refHST).data, numpy.zeros((20, 10)) + 75 + series
                )
            )


class TestDarkCalculationOneFrame(TestCaseQt):
    """Make sure computation of the Dark is correct"""

    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        n_scan = 1
        n_info = 1
        n_xml = 1
        MockEDF.fastMockAcquisition(self.tmp_dir, n_radio=n_scan)
        file_desc = fabio.edfimage.EdfImage(data=numpy.zeros((20, 10)) + 10)

        file_desc.write(os.path.join(self.tmp_dir, "darkend0000.edf"))
        assert len(os.listdir(self.tmp_dir)) is (1 + n_scan + n_info + n_xml)
        self.recons_params = DKRFRP()
        self.darkRef = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "force_sync": True,
                "data": self.tmp_dir,
                "serialize_output_data": False,
            }
        )

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        super().tearDown()

    def testDarkMeanCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEAN

        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "dark.edf")
        self.assertTrue(os.path.isfile(refHST))
        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 10)
        )

    def testDarkMedianCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEDIAN

        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "dark.edf")
        self.assertTrue(os.path.isfile(refHST))

        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 10)
        )


class TestDarkCalculation(TestCaseQt):
    """Make sure computation of the Dark is correct"""

    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        n_scan = 1
        n_xml = 1
        n_info = 1
        MockEDF.fastMockAcquisition(os.path.join(self.tmp_dir), n_radio=n_scan)

        file_desc = fabio.edfimage.EdfImage(data=numpy.zeros((20, 10)))
        file_desc.append_frame(data=(numpy.zeros((20, 10)) + 100))
        file_desc.append_frame(data=(numpy.zeros((20, 10)) + 100))
        file_desc.append_frame(data=(numpy.zeros((20, 10)) + 100))

        file_desc.write(os.path.join(self.tmp_dir, "darkend0000.edf"))
        assert len(os.listdir(self.tmp_dir)) is (1 + n_scan + n_xml + n_info)
        self.recons_params = DKRFRP()
        self.darkRef = DarkRefsTask(
            inputs={
                "data": self.tmp_dir,
                "dark_ref_params": self.recons_params,
                "force_sync": True,
                "serialize_output_data": False,
            }
        )

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        super().tearDown()

    def testDarkMeanCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEAN

        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "dark.edf")
        self.assertTrue(os.path.isfile(refHST))
        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 75)
        )

    def testDarkMedianCalculation(self):
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEDIAN

        self.darkRef.run()
        refHST = os.path.join(self.tmp_dir, "dark.edf")
        self.assertTrue(os.path.isfile(refHST))

        self.assertTrue(
            numpy.array_equal(fabio.open(refHST).data, numpy.zeros((20, 10)) + 100)
        )


class TestDarkAccumulation(TestCaseQt):
    """
    Make sure computation for dark in accumulation are correct
    """

    def setUp(self):
        super().setUp()
        self.dataset = "bone8_1_"
        dataDir = TomwerCIDatasets.get_dataset(
            f"edf_datasets/{self.dataset}",
        )
        self.outputdir = tempfile.mkdtemp()
        shutil.copytree(src=dataDir, dst=os.path.join(self.outputdir, self.dataset))
        self.darkFile = os.path.join(self.outputdir, self.dataset, "dark.edf")
        # create a single 'bone8_1_' to be a valid acquisition directory
        MockEDF.fastMockAcquisition(os.path.join(self.outputdir, self.dataset))
        assert os.path.isfile(self.darkFile)
        with fabio.open(self.darkFile) as dsc:
            self.dark_reference = dsc.data
        # remove dark file
        os.remove(self.darkFile)

        self.recons_params = DKRFRP()
        self.recons_params.flat_calc_method = ReduceMethod.NONE
        self.recons_params.dark_calc_method = ReduceMethod.MEDIAN
        self.recons_params.dark_pattern = "darkend*"
        self.recons_params.dark_prefix = "dark.edf"

    def tearDown(self):
        shutil.rmtree(self.outputdir)
        super().tearDown()

    def testComputation(self):
        """Test data `bone8_1_` from id16b containing dark.edf of reference
        and darkend"""
        dark_ref_process = DarkRefsTask(
            inputs={
                "data": os.path.join(self.outputdir, self.dataset),
                "force_sync": True,
                "serialize_output_data": False,
            }
        )

        dark_ref_process.run()
        self.assertTrue(os.path.isfile(self.darkFile))
        with fabio.open(self.darkFile) as dsc:
            self.computed_dark = dsc.data
        self.assertTrue(numpy.array_equal(self.computed_dark, self.dark_reference))



from __future__ import annotations

import logging
import weakref
import h5py

import numpy

import os
import silx
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.data.DataViewerFrame import DataViewerFrame
from silx.gui.data.DataViews import IMAGE_MODE, DataViewHooks
from silx.gui.dialog.ColormapDialog import ColormapDialog

from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.visualization.reconstructionparameters import ReconstructionParameters
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.resourcemanager import BaseResourceObserver, HDF5VolumeManager
from tomoscan.identifier import VolumeIdentifier

from silx.gui.widgets.WaitingOverlay import WaitingOverlay

_logger = logging.getLogger(__name__)


class _ScanInfo(qt.QWidget):
    """Display information about the reconstruction currently displayed"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        self._scanQLE = qt.QLineEdit("", self)
        self._scanQLE.setReadOnly(True)
        self.layout().addRow("scan", self._scanQLE)

        self._volumeQLE = qt.QLineEdit("", self)
        self._volumeQLE.setReadOnly(True)
        self.layout().addRow("volume", self._volumeQLE)

    def setScan(self, scan: None | TomwerScanBase):
        if scan is None:
            self._scanQLE.clear()
        else:
            assert isinstance(
                scan, TomwerScanBase
            ), f"scan should be an instance of {TomwerScanBase}. Got {type(scan)}"
            self._scanQLE.setText(str(scan))
            self.setVolumes(volumes=scan.latest_vol_reconstructions)

    def setVolumes(self, volumes: tuple[str | VolumeIdentifier | TomwerVolumeBase]):
        """
        Display volume metadata. At the moment only display the identifier of the first reconstructed volume
        """
        self._volumeQLE.clear()
        if len(volumes) > 0:
            volume = volumes[0]
            if not isinstance(volume, TomwerVolumeBase):
                volume = VolumeFactory.create_tomo_object_from_identifier(volume)
            self._volumeQLE.setText(str(volume))
            self._volumeQLE.setToolTip(volume.data_url.file_path())

    def clear(self):
        self.setScan(None)


class _TomoApplicationContext(DataViewHooks):
    """
    Store the context of the application

    It overwrites the DataViewHooks to custom the use of the DataViewer for
    the silx view application.

    - Create a single colormap shared with all the views
    - Create a single colormap dialog shared with all the views
    """

    def __init__(self, parent, settings=None):
        self.__parent = weakref.ref(parent)
        self.__defaultColormap = Colormap(name="gray")
        self.__defaultColormapDialog = None
        self.__settings = settings
        self.__recentFiles = []

    def getSettings(self) -> qt.QSettings:
        """Returns actual application settings."""
        return self.__settings

    def restoreLibrarySettings(self):
        """Restore the library settings, which must be done early"""
        settings = self.__settings
        if settings is None:
            return
        settings.beginGroup("library")
        plotBackend = settings.value("plot.backend", "")
        plotImageYAxisOrientation = settings.value("plot-image.y-axis-orientation", "")
        settings.endGroup()

        if plotBackend != "":
            silx.config.DEFAULT_PLOT_BACKEND = plotBackend
        if plotImageYAxisOrientation != "":
            silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION = (
                plotImageYAxisOrientation
            )

    def restoreSettings(self):
        """Restore the settings of all the application"""
        settings = self.__settings
        if settings is None:
            return
        parent = self.__parent()
        parent.restoreSettings(settings)

        settings.beginGroup("colormap")
        byteArray = settings.value("default", None)
        if byteArray is not None:
            try:
                colormap = Colormap()
                colormap.restoreState(byteArray)
                self.__defaultColormap = colormap
            except Exception:
                _logger.debug("Backtrace", exc_info=True)
        settings.endGroup()

        self.__recentFiles = []
        settings.beginGroup("recent-files")
        for index in range(1, 10 + 1):
            if not settings.contains("path%d" % index):
                break
            filePath = settings.value("path%d" % index)
            self.__recentFiles.append(filePath)
        settings.endGroup()

    def saveSettings(self):
        """Save the settings of all the application"""
        settings = self.__settings
        if settings is None:
            return
        parent = self.__parent()
        parent.saveSettings(settings)

        if self.__defaultColormap is not None:
            settings.beginGroup("colormap")
            settings.setValue("default", self.__defaultColormap.saveState())
            settings.endGroup()

        settings.beginGroup("library")
        settings.setValue("plot.backend", silx.config.DEFAULT_PLOT_BACKEND)
        settings.setValue(
            "plot-image.y-axis-orientation",
            silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION,
        )
        settings.endGroup()

        settings.beginGroup("recent-files")
        for index in range(0, 11):
            key = "path%d" % (index + 1)
            if index < len(self.__recentFiles):
                filePath = self.__recentFiles[index]
                settings.setValue(key, filePath)
            else:
                settings.remove(key)
        settings.endGroup()

    def getRecentFiles(self) -> list[str]:
        """Returns the list of recently opened files.

        The list is limited to the last 10 entries. The newest file path is
        in first.
        """
        return self.__recentFiles

    def pushRecentFile(self, filePath: str):
        """Push a new recent file to the list.

        If the file is duplicated in the list, all duplications are removed
        before inserting the new filePath.

        If the list becan bigger than 10 items, oldest paths are removed.

        :param filePath: File path to push
        """
        # Remove old occurencies
        self.__recentFiles[:] = (f for f in self.__recentFiles if f != filePath)
        self.__recentFiles.insert(0, filePath)
        while len(self.__recentFiles) > 10:
            self.__recentFiles.pop()

    def clearRencentFiles(self):
        """Clear the history of the recent files."""
        self.__recentFiles[:] = []

    def getColormap(self, view) -> Colormap:
        """Returns a default colormap.

        Override from DataViewHooks
        """
        if self.__defaultColormap is None:
            self.__defaultColormap = Colormap(name="viridis")
        return self.__defaultColormap

    def getColormapDialog(self, view) -> ColormapDialog:
        """Returns a shared color dialog as default for all the views.

        Override from DataViewHooks
        """
        if self.__defaultColormapDialog is None:
            parent = self.__parent()
            if parent is None:
                return None
            dialog = ColormapDialog(parent=parent)
            dialog.setModal(False)
            self.__defaultColormapDialog = dialog
        return self.__defaultColormapDialog


class VolumeViewer(qt.QMainWindow, BaseResourceObserver):
    def __init__(self, parent):
        super().__init__(parent)

        self._volume_loaded_in_background = (None, None)
        # store the volume loaded in the background and the thread used for it as (volume_identifier, thread)
        self._centralWidget = DataViewerFrame(parent=self)
        self.__context = _TomoApplicationContext(self)
        self._centralWidget.setGlobalHooks(self.__context)
        self._centralWidget.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        self.setCentralWidget(self._centralWidget)
        # waiter overlay to notify user loading is on-going
        self._waitingOverlay = WaitingOverlay(self._centralWidget)
        self._waitingOverlay.setIconSize(qt.QSize(30, 30))
        self._waitingOverlay.hide()

        # display scan information when possible
        self._infoWidget = _ScanInfo(parent=self)

        # top level dock widget to display information regarding the scan
        # and volume
        self._dockInfoWidget = qt.QDockWidget(parent=self)
        self._dockInfoWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._dockInfoWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._dockInfoWidget.setWidget(self._infoWidget)
        self.addDockWidget(qt.Qt.TopDockWidgetArea, self._dockInfoWidget)

        # add dock widget for reconstruction parameters
        self._reconsInfoDockWidget = qt.QDockWidget(parent=self)
        self._reconsWidgetScrollArea = qt.QScrollArea(self)
        self._reconsWidgetScrollArea.setWidgetResizable(True)
        self._reconsWidget = ReconstructionParameters(self)
        self._reconsWidgetScrollArea.setWidget(self._reconsWidget)
        self._reconsInfoDockWidget.setWidget(self._reconsWidgetScrollArea)
        self._reconsInfoDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._reconsInfoDockWidget)

        self._h5_file = None
        """pointer to the hdf5 file since we want to set the HDF5Dataset for
        loading data on the fly and avoid loading everything into memory for
        hdf5."""
        self.__last_mode = None
        # register the viewer as observer of the HDF5 volume (handle resource conflict)
        HDF5VolumeManager.register(self)

    def _close_h5_file(self):
        if self._h5_file is not None:
            self._h5_file.close()
        self._h5_file = None

    def close(self):
        HDF5VolumeManager.unregister(self)
        self._close_h5_file()
        super().close()

    def release_resource(self, resource: str):
        if self._h5_file is not None and resource in (
            self._h5_file.filename,
            os.path.abspath(self._h5_file.filename),
        ):
            _logger.warning(
                f"{resource} about to be overwrite. Closing resource and cleaning plot."
            )
            self._close_h5_file()
            self.clear()

    def setScan(self, scan):
        self.clear()
        if scan is None:
            return

        elif len(scan.latest_vol_reconstructions) == 0:
            _logger.warning(f"No reconstructed volume for {scan}")
            self._infoWidget.setScan(scan)
        else:
            self._set_volumes(volumes=scan.latest_vol_reconstructions)
            self._infoWidget.setScan(scan)

    def setVolume(self, volume):
        self.clear(clear_infos=False)
        if volume is None:
            return
        self._set_volumes(volumes=(volume,))
        self._infoWidget.setVolumes((volume,))

    def _get_data_volume(self, volume: TomwerVolumeBase):
        """
        load the data of the requested volume.
        :return: (data: str | None, state: str) state can be "loaded", "loading" or "failed"
        """
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError(
                f"volume is expected to be an instance of {TomwerVolumeBase}. Not {type(volume)}"
            )

        self._close_h5_file()
        if isinstance(volume, HDF5Volume):
            try:
                self._h5_file = h5py.File(volume.data_url.file_path(), mode="r")
            except OSError:
                self._h5_file = h5py.File(
                    volume.data_url.file_path(),
                    mode="r",
                    libver="latest",
                    swmr=True,
                )
            if volume.data_url.data_path() in self._h5_file:
                data = self._h5_file[volume.data_url.data_path()]
                state = "loaded"
            else:
                data = None
                state = "failed"
        elif volume.data is not None:
            data = volume.data
            state = "loaded"
        else:
            volume_id, _ = self._volume_loaded_in_background
            if volume_id == volume.get_identifier().to_str():
                # special case if the user send several time the volume which is currently loading
                # in this case we just want to ignore the request to avoid reloading the volume
                # can happen in the case of a large volume that take some time to be loaded.
                pass
            else:
                _logger.warning(
                    "Attempt to set a non HDF5 volume to the viewer. This requires to load all the data in memory. This can take a while"
                )
                self._stopExistingLoaderThread()
                self._loadAndDisplay(volume)
            state = "loading"
            data = None

        return data, state

    def _loaderThreadFinished(self):
        """Callback activated when a VolumeLoader thread is finished"""
        sender = self.sender()
        if not isinstance(sender, VolumeLoader):
            raise TypeError("sender is expected to be a VolumeLoader")

        self._stopExistingLoaderThread()
        if sender.volume.data is None:
            _logger.error(f"Failed to load volume {sender.volume.get_identifier()}")
        elif sender.volume is not None:
            self.setVolume(sender.volume)

    def _loadAndDisplay(self, volume):
        """Load a thread and add a callback when loading is done"""
        loader_thread = VolumeLoader(volume=volume)
        self._volume_loaded_in_background = (
            volume.get_identifier().to_str(),
            loader_thread,
        )
        loader_thread.finished.connect(self._loaderThreadFinished)
        loader_thread.start()

    def _stopExistingLoaderThread(self):
        """Will stop any existing loader thread. Make sure we load one volume at most at the time"""
        _, loader_thread = self._volume_loaded_in_background
        if loader_thread is not None:
            loader_thread.finished.disconnect(self._loaderThreadFinished)
            if loader_thread.isRunning():
                loader_thread.quit()
            self._volume_loaded_in_background = (None, None)

    def _set_volumes(self, volumes: tuple):
        self.clear()
        # for now handle a single volume
        if len(volumes) == 0:
            pass
        else:
            if len(volumes) > 1:
                _logger.warning(
                    "Only one volume can be displayed. Will display the first one"
                )
            volume = volumes[0]
            if isinstance(volume, (str, VolumeIdentifier)):
                volume = VolumeFactory.create_tomo_object_from_identifier(volume)
            elif isinstance(volume, TomwerVolumeBase):
                pass
            else:
                raise TypeError(
                    f"Volume should be an instance of a Volume, a VolumeIdentifier or a string refering to a VolumeIdentifier. {type(volume)} provided"
                )

            # warning: load metadata before data because can get some conflict with the HDF5 reader flag if done after
            try:
                # warning: limitation expected for .vol as it gets two configuration file. The default one is vol.info and does not contains
                # any of the metadata 'distance', 'pixel size'... but it is here for backward compatiblity
                self._reconsWidget.setVolumeMetadata(
                    volume.metadata or volume.load_metadata()
                )
            except Exception as e:
                _logger.info(
                    f"Unable to set reconstruction parameters from {volume.data_url}. Not handled for pyhst reconstructions. Error is {e}"
                )

            data, state = self._get_data_volume(volume)
            # set volume dataset
            if state == "loading":
                self._waitingOverlay.show()
            elif state == "loaded":
                self._waitingOverlay.hide()
                if data is not None:
                    self._set_volume(data)

            elif state == "failed":
                _logger.warning(
                    f"Failed to load data from {volume.get_identifier().to_str()}"
                )
                return

    def _set_volume(self, volume: numpy.ndarray | h5py.Dataset):
        self._centralWidget.setData(volume)
        self._centralWidget.setDisplayMode(self.__last_mode or IMAGE_MODE)

    def clear(self, clear_infos=True):
        if self._centralWidget.displayMode():
            self.__last_mode = self._centralWidget.displayMode()
        self._close_h5_file()
        if clear_infos:
            self._infoWidget.clear()
        self._centralWidget.setData(None)
        # if clear stop loading any volume
        self._stopExistingLoaderThread()

    def sizeHint(self):
        return qt.QSize(600, 600)


class VolumeLoader(qt.QThread):
    """
    simple thread that load a volume in memory
    """

    def __init__(self, volume: TomwerVolumeBase) -> None:
        super().__init__()
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError()
        self.__volume = volume

    def run(self):
        self.__volume.load_data(store=True)

    @property
    def volume(self):
        return self.__volume



# coding: utf-8
from __future__ import annotations


import gc

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.visualization.VolumeViewerOW import VolumeViewerOW
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestVolumeViewerOW(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        super().setUp()
        self._window = VolumeViewerOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        del self._window._window
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)



from __future__ import annotations

from silx.gui import qt

from tomwer.core.process.reconstruction.axis.mode import AxisMode, AXIS_MODE_METADATAS
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.axis.params import AxisCalculationInput
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel, QSpinBoxIgnoreWheel
from tomwer.core.process.reconstruction.axis.params import (
    DEFAULT_CMP_N_SUBSAMPLING_Y,
    DEFAULT_CMP_OVERSAMPLING,
    DEFAULT_CMP_TAKE_LOG,
    DEFAULT_CMP_THETA,
)


class AxisOptionsWidget(qt.QWidget):
    """
    GUI to define (advanced) option of the AxisTask

    Used as a tab of the AxisSettingsTabWidget
    """

    sigChanged = qt.Signal()
    """Emit when the options changed"""

    def __init__(self, parent, axis_params):
        qt.QWidget.__init__(self, parent=parent)
        assert isinstance(axis_params, QAxisRP)
        self._axis_params = axis_params
        self.setLayout(qt.QVBoxLayout())

        # cor_options
        self._corOptsWidget = qt.QWidget(self)
        self._corOptsWidget.setLayout(qt.QFormLayout())
        self._corOpts = qt.QLineEdit(self)
        self._corOpts.setToolTip(
            "Options for methods finding automatically the rotation axis position. 'side', 'near_pos' and 'near_width' are already provided by dedicated interface. The parameters are separated by commas and passed as 'name=value'. Mind the semicolon separator (;)."
        )
        self._corOpts.setPlaceholderText("low_pass=1; high_pass=20")
        self._corOptsWidget.layout().addRow("cor advanced options", self._corOpts)
        self.layout().addWidget(self._corOptsWidget)

        # padding option
        self._padding_widget = qt.QGroupBox("padding mode")
        self._padding_widget.setCheckable(True)
        self.layout().addWidget(self._padding_widget)
        self._padding_widget.setLayout(qt.QHBoxLayout())

        self._qbPaddingMode = QComboBoxIgnoreWheel(self._padding_widget)
        for _mode in (
            "constant",
            "edge",
            "linear_ramp",
            "maximum",
            "mean",
            "median",
            "minimum",
            "reflect",
            "symmetric",
            "wrap",
        ):
            self._qbPaddingMode.addItem(_mode)
        def_index = self._qbPaddingMode.findText("edge")
        self._qbPaddingMode.setCurrentIndex(def_index)
        self._padding_widget.layout().addWidget(self._qbPaddingMode)

        # define common options
        self._commonOpts = qt.QWidget(parent=self)
        self._commonOpts.setLayout(qt.QFormLayout())

        self._qcbDataMode = qt.QComboBox(parent=self)
        for data_mode in AxisCalculationInput:
            # paganin is not managed for sinogram
            self._qcbDataMode.addItem(data_mode.name(), data_mode)
        self._qcbDataMode.hide()

        self.layout().addWidget(self._commonOpts)

        # composite method advanced options
        self._compositeOptsGroup = CompositeOptsGroup(
            parent=self, axis_params=axis_params
        )
        self.layout().addWidget(self._compositeOptsGroup)

        # set up
        self.setCalculationInputType(self._axis_params.calculation_input_type)
        self._compositeOptsGroup.setVisible(
            self._axis_params.mode in (AxisMode.near, AxisMode.composite_coarse_to_fine)
        )

        # connect signal / slot
        self._corOpts.editingFinished.connect(self._updateAdvancedCorOptions)
        self._qcbDataMode.currentIndexChanged.connect(self._updateInputType)
        self._axis_params.sigChanged.connect(self._axis_params_changed)
        self._qbPaddingMode.currentIndexChanged.connect(self._paddingModeChanged)
        self._padding_widget.toggled.connect(self._paddingModeChanged)
        self._compositeOptsGroup.sigChanged.connect(self.sigChanged)

    def _axis_params_changed(self):
        with block_signals(self):
            # update according to AxisCalculationInput
            index = self._qcbDataMode.findText(
                self._axis_params.calculation_input_type.name()
            )
            if index >= 0:
                self._qcbDataMode.setCurrentIndex(index)
            # update advanced cor options visibility (not relevant if mode is manual or read)
            axis_mode = self._axis_params.mode
            self._corOptsWidget.setVisible(
                axis_mode
                not in (
                    AxisMode.manual,
                    AxisMode.read,
                )
            )
            # update cor options value
            self.setCorOptions(self._axis_params.extra_cor_options)
            self.setPaddingMode(self._axis_params.padding_mode)
            self._padding_widget.setVisible(
                AXIS_MODE_METADATAS[axis_mode].allows_padding
            )

    def _updateInputType(self, *arg, **kwargs):
        self._axis_params.calculation_input_type = self.getCalculationInputType()
        self.sigChanged.emit()

    def _paddingModeChanged(self, *args, **kwargs):
        self._axis_params.padding_mode = self.getPaddingMode()
        self.sigChanged.emit()

    def getPaddingMode(self):
        if self._padding_widget.isChecked():
            return self._qbPaddingMode.currentText()
        else:
            return None

    def setPaddingMode(self, mode):
        index = self._qbPaddingMode.findText(mode)
        if index >= 0:
            self._qbPaddingMode.setCurrentIndex(index)
        self._paddingModeChanged()

    def _updateAdvancedCorOptions(self, *args, **kwargs):
        self._axis_params.extra_cor_options = self.getCorOptions()
        self.sigChanged.emit()

    def getCalculationInputType(self, *arg, **kwargs):
        return AxisCalculationInput.from_value(self._qcbDataMode.currentText())

    def setCalculationInputType(self, calculation_type):
        calculation_type = AxisCalculationInput.from_value(calculation_type)
        index_dm = self._qcbDataMode.findText(calculation_type.name())
        self._qcbDataMode.setCurrentIndex(index_dm)

    def setAxisParams(self, axis):
        self._axis_params = axis
        with block_signals(self):
            index = self._qcbDataMode.findText(axis.calculation_input_type.name())
            self._qcbDataMode.setCurrentIndex(index)
            self._compositeOptsGroup.setAxisParams(axis)

    def getCorOptions(self):
        return self._corOpts.text()

    def setCorOptions(self, opts: str | None):
        with block_signals(self._axis_params):
            self._corOpts.clear()
            if opts:
                self._corOpts.setText(opts)
                self._updateAdvancedCorOptions()

    def setMode(self, mode: AxisMode):
        composite_opts_visible = AxisMode.from_value(mode) in (
            AxisMode.composite_coarse_to_fine,
            AxisMode.near,
        )
        self._compositeOptsGroup.setVisible(composite_opts_visible)


class CompositeOptsGroup(qt.QGroupBox):
    """Group box dedicated to the composite algorithms"""

    sigChanged = qt.Signal()
    """Emit when the options changed"""

    def __init__(self, title="composite options", parent=None, axis_params=None):
        self._axis_params = axis_params
        super().__init__(title, parent)
        ## options for the composite mode
        self.setLayout(qt.QFormLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._thetaSB = QSpinBoxIgnoreWheel(self)
        self._thetaSB.setRange(0, 360)
        self._thetaSB.setValue(DEFAULT_CMP_THETA)
        self._thetaSB.setToolTip("a radio will be picked each theta degrees")
        self._thetaLabel = qt.QLabel("angle interval (in degree)", self)
        self._thetaLabel.setToolTip(
            "algorithm will take one projection each 'angle interval'. Also know as 'theta'"
        )
        self.layout().addRow(self._thetaLabel, self._thetaSB)

        self._oversamplingSB = QSpinBoxIgnoreWheel(self)
        self._oversamplingSB.setRange(1, 999999)
        self._oversamplingSB.setValue(DEFAULT_CMP_OVERSAMPLING)
        self._oversamplingSB.setToolTip("sinogram oversampling")
        self.layout().addRow("oversampling", self._oversamplingSB)

        self._nearWidthSB = QSpinBoxIgnoreWheel(self)
        self._nearWidthSB.setRange(1, 999999)
        self._nearWidthSB.setValue(0)
        self._nearWidthSB.setToolTip("position to be used with near option")
        self._nearWidthLabel = qt.QLabel("near width", self)
        self._nearWidthLabel.setToolTip("position to be used with near option")
        self.layout().addRow(self._nearWidthLabel, self._nearWidthSB)

        self._subsamplingYSB = QSpinBoxIgnoreWheel(self)
        self._subsamplingYSB.setRange(1, 999999)
        self._subsamplingYSB.setValue(DEFAULT_CMP_N_SUBSAMPLING_Y)
        self._subsamplingYSB.setToolTip("sinogram number of subsampling along y")
        self.layout().addRow("n_subsampling_y", self._subsamplingYSB)

        self._takeLogCB = qt.QCheckBox(self)
        self._takeLogCB.setToolTip("Take logarithm")
        self._takeLogCB.setChecked(DEFAULT_CMP_TAKE_LOG)
        self._takeTheLogLabel = qt.QLabel("linearisation (-log(I/I0))")
        self._takeTheLogLabel.setToolTip(
            "take (-log(I/I0)) as input. Also know as 'take_log' option"
        )
        self.layout().addRow(self._takeTheLogLabel, self._takeLogCB)

        # connect signal / slot
        self._thetaSB.valueChanged.connect(self._changed)
        self._oversamplingSB.valueChanged.connect(self._changed)
        self._subsamplingYSB.valueChanged.connect(self._changed)
        self._nearWidthSB.valueChanged.connect(self._changed)
        self._takeLogCB.toggled.connect(self._changed)

    def setAxisParams(self, axis_params):
        with block_signals(self):
            self.setConfiguration(axis_params.composite_options)
        self._axis_params = axis_params

    def _changed(self):
        if self._axis_params is not None:
            self._axis_params.composite_options = self.getConfiguration()
        self.sigChanged.emit()

    def getTheta(self) -> int:
        return self._thetaSB.value()

    def setTheta(self, theta: int) -> None:
        self._thetaSB.setValue(theta)

    def getOversampling(self) -> int:
        return self._oversamplingSB.value()

    def setOversampling(self, oversampling: int) -> None:
        self._oversamplingSB.setValue(oversampling)

    def getNearWidth(self) -> int:
        return self._nearWidthSB.value()

    def setNearWidth(self, value) -> int:
        return self._nearWidthSB.setValue(value)

    def getSubSamplingY(self) -> int:
        return self._subsamplingYSB.value()

    def setSubSamplingY(self, subsampling: int) -> None:
        self._subsamplingYSB.setValue(subsampling)

    def getTakeLog(self) -> bool:
        return self._takeLogCB.isChecked()

    def setTakeLog(self, log: bool) -> None:
        self._takeLogCB.setChecked(log)

    def getConfiguration(self) -> dict:

        return {
            "theta": self.getTheta(),
            "oversampling": self.getOversampling(),
            "n_subsampling_y": self.getSubSamplingY(),
            "take_log": self.getTakeLog(),
            "near_width": self.getNearWidth(),
        }

    def setConfiguration(self, opts: dict) -> None:
        if not isinstance(opts, dict):
            raise TypeError("opts should be an instance of dict")
        # theta
        theta = opts.get("theta", None)
        if theta is not None:
            self.setTheta(theta=theta)
        # oversampling
        oversampling = opts.get("oversampling", None)
        if oversampling is not None:
            self.setOversampling(oversampling)
        # n subsampling y
        n_subsampling_y = opts.get("n_subsampling_y", None)
        if n_subsampling_y is not None:
            self.setSubSamplingY(n_subsampling_y)
        # near_width
        near_width = opts.get("near_width", None)
        if near_width is not None:
            self.setNearWidth(near_width)
        # take log
        take_log = opts.get("take_log", None)
        if take_log is not None:
            self.setTakeLog(take_log)



from __future__ import annotations

import logging

from silx.gui import qt

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Output, OWBaseWidget

from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.gui.cluster.slurm import SlurmSettingsWindow
from tomwer.core.settings import SlurmSettingsMode

_logger = logging.getLogger(__name__)


class SlurmClusterOW(OWBaseWidget, openclass=True):
    """
    Orange widget to define a slurm cluster as input of other
    widgets (based on nabu for now)
    """

    name = "slurm cluster"
    id = "orange.widgets.tomwer.cluster.SlurmClusterOW.SlurmClusterOW"
    description = "Let the user configure the cluster to be used."
    icon = "icons/slurm.svg"
    priority = 20
    keywords = ["tomography", "tomwer", "slurm", "cluster", "configuration"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _ewoks_default_inputs = Setting(dict())
    slurm_mode_name = Setting(SlurmSettingsMode.GENERIC.name)

    class Outputs:
        config_out = Output(name="cluster_config", type=SlurmClusterConfiguration)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = SlurmSettingsWindow(parent=self)
        self._widget.setWindowFlags(qt.Qt.Widget)
        layout.addWidget(self._widget)

        if self._ewoks_default_inputs != {}:
            self._widget.setConfiguration(self._ewoks_default_inputs)

        try:
            mode = SlurmSettingsMode[self.slurm_mode_name]
            self._widget.setMode(mode)
        except Exception:
            _logger.exception(
                f"Could not restore saved SlurmSettingsMode {self.slurm_mode_name!r}"
            )

        # trigger the signal to avoid any user request
        self.Outputs.config_out.send(self.getConfiguration())

        # connect signal / slot
        self._widget.sigConfigChanged.connect(self._configurationChanged)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("_static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def _configurationChanged(self):
        slurmClusterconfiguration = self.getConfiguration()
        self.Outputs.config_out.send(slurmClusterconfiguration)
        self._ewoks_default_inputs = slurmClusterconfiguration.to_dict()
        self.slurm_mode_name = self._widget.mode().name

    def getConfiguration(self):
        return self._widget.getSlurmClusterConfiguration()



from __future__ import annotations

import numpy
import logging
from silx.gui import qt
from tomwer.core.utils.char import BETA_CHAR, DELTA_CHAR, MU_CHAR

_logger = logging.getLogger(__name__)


class ReconstructionParameters(qt.QWidget):
    """
    display reconstruction parameters of a volume
    """

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        # method
        self._methodQLE = qt.QLineEdit("", self)
        self.layout().addRow("Method", self._methodQLE)
        self._methodQLE.setReadOnly(True)
        # paganin
        self._paganinQLE = qt.QLineEdit("", self)
        self._paganinQLE.setReadOnly(True)
        self.layout().addRow("Phase method", self._paganinQLE)
        # delta beta
        self._deltaBetaLabel = qt.QLabel(" / ".join((DELTA_CHAR, BETA_CHAR)), self)
        self._deltaBetaQLE = qt.QLineEdit("", self)
        self._deltaBetaQLE.setReadOnly(True)
        self.layout().addRow(self._deltaBetaLabel, self._deltaBetaQLE)
        # sample_detector distance
        self._sampleDetectorDistanceInMeterQLE = qt.QLineEdit("", self)
        self._sampleDetectorDistanceInMeterQLE.setReadOnly(True)
        self.layout().addRow(
            "Sample-detector distance (m)", self._sampleDetectorDistanceInMeterQLE
        )
        # pixel size
        self._voxelSizeInMicronQLE = qt.QLineEdit("", self)
        self._voxelSizeInMicronQLE.setReadOnly(True)
        self.layout().addRow(f"Voxel size ({MU_CHAR}m)", self._voxelSizeInMicronQLE)
        # cor
        self._corQLE = qt.QLineEdit("", self)
        self._corQLE.setReadOnly(True)
        self.layout().addRow("CoR (absolute)", self._corQLE)
        # padding type
        self._paddingTypeQLE = qt.QLineEdit("", self)
        self._paddingTypeQLE.setReadOnly(True)
        self.layout().addRow("Padding type", self._paddingTypeQLE)
        # half tomo
        self._halfTomoCB = qt.QCheckBox("", self)
        self._halfTomoCB.setEnabled(False)
        self.layout().addRow("Half tomo", self._halfTomoCB)
        # fbp filter type
        self._fbpFilterQLE = qt.QLineEdit("", self)
        self._fbpFilterQLE.setReadOnly(True)
        self.layout().addRow("FBP filter", self._fbpFilterQLE)
        # log min clip
        self._minLogClipQLE = qt.QLineEdit("", self)
        self._minLogClipQLE.setReadOnly(True)
        self.layout().addRow("Log min clip", self._minLogClipQLE)
        # log max clip
        self._maxLogClipQLE = qt.QLineEdit("", self)
        self._maxLogClipQLE.setReadOnly(True)
        self.layout().addRow("Log max clip", self._maxLogClipQLE)
        # sino normalization & normalization file
        self._sinonormalizationQLE = qt.QLabel("", self)
        self.layout().addRow("Sinogram normalization", self._sinonormalizationQLE)
        self._sinonormalizationFileQLE = qt.QLabel("", self)
        self.layout().addRow(
            "Sinogram normalization file", self._sinonormalizationFileQLE
        )
        # software version
        self._softwareVersionQLE = qt.QLabel("", self)
        self.layout().addRow("Software version", self._softwareVersionQLE)

        # Connect signal for paganinQLE
        self._paganinQLE.textChanged.connect(self._updateDeltaBetaVisibility)

    def _updateDeltaBetaVisibility(self):
        phase_method = self._paganinQLE.text()
        display_delta_beta = phase_method in ("paganin", "CTF")
        self._deltaBetaLabel.setVisible(display_delta_beta)
        self._deltaBetaQLE.setVisible(display_delta_beta)

    def setVolumeMetadata(self, metadata: dict | None):
        if metadata is None:
            metadata = {}
        elif not isinstance(metadata, dict):
            raise TypeError(f"url should be a {dict}. {type(metadata)} provided")

        for func in (
            self._setMethod,
            self._setPhaseMethod,
            self._setDeltaBeta,
            self._setDistance,
            self._setVoxelSize,
            self._setCor,
            self._setPaddingType,
            self._setHalfTomo,
            self._setFBPFilter,
            self._setMinLogClip,
            self._setMaxLogClip,
            self._setSinoNormalization,
            self._setSoftwareVersion,
        ):
            try:
                func(metadata)
            except Exception as e:
                _logger.warning(f"Fail update when call {func}. Error is", e)

    def _setMethod(self, metadata: dict):
        method = (
            metadata.get("nabu_config", {}).get("reconstruction", {}).get("method", "")
        )
        self._methodQLE.setText(method)

    def _setPhaseMethod(self, metadata: dict):
        phase_method = (
            metadata.get("nabu_config", {}).get("phase", {}).get("method", "")
        )
        # note: pahse method is expected to be in ("", "paganin", "CTF")
        self._paganinQLE.setText(phase_method)

    def _setDeltaBeta(self, metadata: dict):
        delta_beta = (
            metadata.get("nabu_config", {}).get("phase", {}).get("delta_beta", "")
        )
        self._deltaBetaQLE.setText(str(delta_beta))

    def _setDistance(self, metadata: dict):
        distance_cm = (
            metadata.get("processing_options", {})
            .get("phase", {})
            .get("distance_cm", None)
        ) or metadata.get("processing_options", {}).get("reconstruction", {}).get(
            "sample_detector_dist", None
        )
        if distance_cm not in (None, "", "None"):
            distance_m = float(distance_cm) / 100.0
            distance_m = f"{distance_m:.5}"
        else:
            distance_m = ""
        self._sampleDetectorDistanceInMeterQLE.setText(distance_m)

    def _setVoxelSize(self, metadata: dict):
        # voxel size can be stored as pixel size (old version) or voxel size (new version)
        recons_params = metadata.get("processing_options", {}).get("reconstruction", {})
        voxel_size_cm = recons_params.get("voxel_size_cm", [None] * 3)
        # back compatibility when voxel was a scalar ( ~ nabu 2023 ?)
        if numpy.isscalar(voxel_size_cm):
            voxel_size_cm = [voxel_size_cm] * 3

        # now voxel size is expected to be a tuple of three elements
        if voxel_size_cm is not None:

            def clean_voxel_value(value):
                if isinstance(value, str):
                    for char_to_ignore in (" ", "(", ")", "[", "]"):
                        value = value.replace(char_to_ignore, "")
                return value

            voxel_size_cm = [clean_voxel_value(value) for value in voxel_size_cm]

        else:
            # backward compatibility with old volume
            voxel_size_cm = recons_params.get("pixel_size_cm", [None] * 3)

        voxel_size_cm = filter(None, voxel_size_cm)

        def cast_voxel_value(value_cm: float):
            value_micron = value_cm * 10_000.0
            return f"{float(value_micron):.2}"

        voxel_size_micron = [cast_voxel_value(value) for value in voxel_size_cm]
        self._voxelSizeInMicronQLE.setText("x".join(voxel_size_micron))

    def _setCor(self, metadata: dict):
        cor = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("rotation_axis_position", None)
        )
        if cor not in (None, "None", "none"):
            cor = f"{float(cor):.2f}"
        else:
            cor = None
        self._corQLE.setText(cor if cor is not None else "")

    def _setPaddingType(self, metadata: dict):
        padding_type = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("padding_type", "")
        )
        self._paddingTypeQLE.setText(str(padding_type))

    def _setHalfTomo(self, metadata: dict):
        enable_halftomo = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("enable_halftomo", False)
        )
        self._halfTomoCB.setChecked(enable_halftomo in ("True", "true", True, 1, "1"))

    def _setFBPFilter(self, metadata: dict):
        fbp_filter_type = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("fbp_filter_type", "")
        )
        self._fbpFilterQLE.setText(str(fbp_filter_type))

    def _setMinLogClip(self, metadata: dict):
        log_min_clip = (
            metadata.get("processing_options", {})
            .get("take_log", {})
            .get("log_min_clip", "")
        )
        self._minLogClipQLE.setText(str(log_min_clip))

    def _setMaxLogClip(self, metadata: dict):
        log_max_clip = (
            metadata.get("processing_options", {})
            .get("take_log", {})
            .get("log_max_clip", "")
        )
        self._maxLogClipQLE.setText(str(log_max_clip))

    def _setSinoNormalization(self, metadata: dict):
        norm_method = (
            metadata.get("processing_options", {})
            .get("sino_normalization", {})
            .get("method", "")
        )
        sino_normalization_file = (
            metadata.get("nabu_config", {})
            .get("preproc", {})
            .get("sino_normalization_file", "")
        )

        self._sinonormalizationQLE.setText(norm_method)
        self._sinonormalizationFileQLE.setText(sino_normalization_file)
        self._sinonormalizationFileQLE.setToolTip(sino_normalization_file)
        self._sinonormalizationQLE.setVisible(norm_method is not None)
        self._sinonormalizationFileQLE.setVisible(sino_normalization_file is not None)

    def _setSoftwareVersion(self, metadata: dict):
        # for HDF5 the version is part of the metadata[version] section. For other this is part of metadata[process_info]
        software_version = metadata.get("version", None) or metadata.get(
            "process_info", {}
        ).get("nabu_version", None)
        if software_version is None:
            software_version = ""

        software = "nabu"
        self._softwareVersionQLE.setText(f"{software} ({software_version})")



# coding: utf-8
from __future__ import annotations


import pathlib
from datetime import datetime
from processview.core.dataset import Dataset, DatasetIdentifier
from tomoscan.esrf.identifier.jp2kidentifier import (
    JP2KVolumeIdentifier as _JP2KVolumeIdentifier,
)
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume as _JP2KVolume

from tomwer.core.volume.volumebase import TomwerVolumeBase


class JP2KVolumeIdentifier(_JP2KVolumeIdentifier, DatasetIdentifier):
    """Identifier specific to JP2K volume"""

    def __init__(self, object, folder, file_prefix, metadata=None):
        super().__init__(object, folder, file_prefix)
        DatasetIdentifier.__init__(self, JP2KVolume.from_identifier, metadata=metadata)

    @staticmethod
    def from_str(identifier):
        return _JP2KVolumeIdentifier._from_str_to_single_frame_identifier(
            identifier=identifier,
            SingleFrameIdentifierClass=JP2KVolumeIdentifier,
            ObjClass=JP2KVolume,
        )

    def long_description(self) -> str:
        """used for processview header tooltip for now"""
        return self.to_str()


class JP2KVolume(_JP2KVolume, TomwerVolumeBase, Dataset):
    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, JP2KVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {JP2KVolumeIdentifier}"
            )
        return JP2KVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    def get_identifier(self) -> JP2KVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")

        try:
            stat = pathlib.Path(self.url.file_path()).stat()
        except Exception:
            stat = None

        return JP2KVolumeIdentifier(
            object=self,
            folder=self.url.file_path(),
            file_prefix=self._volume_basename,
            metadata={
                "name": self.url.file_path(),
                "creation_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
                "modification_time": (
                    datetime.fromtimestamp(stat.st_ctime) if stat else None
                ),
            },
        )



from silx.gui import qt

from tomwer.gui.edit.nxtomoeditor import NXtomoEditor as _NXtomoEditor


class NXtomoMetadataViewer(_NXtomoEditor):
    """
    class to display metadata of a NXtomo.
    inherit from the `NXtomoEditor` and make sure not edition is possible
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        for widget in self.getEditableWidgets():
            if isinstance(widget, (qt.QComboBox, qt.QCheckBox)):
                widget.setEnabled(False)
            else:
                widget.setReadOnly(True)



# coding: utf-8
# This file come from the django project (https://github.com/django/django) and
# distributed under the BSD license
###########################################################################
# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################


__author__ = "Dhruv Govil"
__copyright__ = "Copyright 2016, Dhruv Govil"
__credits__ = [
    "Dhruv Govil",
    "John Hood",
    "Jason Viloria",
    "Adric Worley",
    "Alex Widener",
]
__license__ = "MIT"
__version__ = "1.1.1"
__maintainer__ = "Dhruv Govil"
__email__ = "dhruvagovil@gmail.com"
__status__ = "Beta"
__date__ = "09/01/2018"

import inspect
import weakref
from functools import partial


class Signal(object):
    """
    The Signal is the core object that handles connection and emission .

    .. warning: This intends to work only for Signal/Slot executed in the same
                thread.
    """

    def __init__(self, *arg_types):
        super(Signal, self).__init__()
        self._block = False
        self._slots = []
        self._arguments_types = arg_types

    def emit(self, *args, **kwargs):
        """
        Calls all the connected slots with the provided args and kwargs unless block is activated
        """

        if self._block:
            return

        for slot in self._slots:
            if not slot:
                continue
            elif isinstance(slot, partial):
                slot()
            elif isinstance(slot, weakref.WeakKeyDictionary):
                # For class methods, get the class object and call the method accordingly.
                for obj, method in slot.items():
                    method(obj, *args, **kwargs)
            elif isinstance(slot, weakref.ref):
                # If it's a weakref, call the ref to get the instance and then call the func
                # Don't wrap in try/except so we don't risk masking exceptions from the actual func call
                if slot() is not None:
                    slot()(*args, **kwargs)
            else:
                # Else call it in a standard way. Should be just lambdas at this point
                slot(*args, **kwargs)

    def connect(self, slot):
        """
        Connects the signal to any callable object
        """
        if not callable(slot):
            raise ValueError(
                "Connection to non-callable '%s' object failed"
                % slot.__class__.__name__
            )

        if isinstance(slot, partial) or "<" in slot.__name__:
            # If it's a partial or a lambda. The '<' check is the only py2 and py3 compatible way I could find
            if slot not in self._slots:
                self._slots.append(slot)
        elif inspect.ismethod(slot):
            # Check if it's an instance method and store it with the instance as the key
            slotSelf = slot.__self__
            slotDict = weakref.WeakKeyDictionary()
            slotDict[slotSelf] = slot.__func__
            if slotDict not in self._slots:
                self._slots.append(slotDict)
        else:
            # If it's just a function then just store it as a weakref.
            newSlotRef = weakref.ref(slot)
            if newSlotRef not in self._slots:
                self._slots.append(newSlotRef)

    def disconnect(self, slot):
        """
        Disconnects the slot from the signal
        """
        if not callable(slot):
            return

        if inspect.ismethod(slot):
            # If it's a method, then find it by its instance
            slotSelf = slot.__self__
            for s in self._slots:
                if (
                    isinstance(s, weakref.WeakKeyDictionary)
                    and (slotSelf in s)
                    and (s[slotSelf] is slot.__func__)
                ):
                    self._slots.remove(s)
                    break
        elif isinstance(slot, partial) or "<" in slot.__name__:
            # If it's a partial or lambda, try to remove directly
            try:
                self._slots.remove(slot)
            except ValueError:
                pass
        else:
            # It's probably a function, so try to remove by weakref
            try:
                self._slots.remove(weakref.ref(slot))
            except ValueError:
                pass

    def clear(self):
        """Clears the signal of all connected slots"""
        self._slots = []

    def block(self, isBlocked):
        """Sets blocking of the signal"""
        self._block = bool(isBlocked)



# coding: utf-8
from __future__ import annotations


import logging
import time
from collections import namedtuple
from datetime import datetime
from multiprocessing import Lock, Manager

from silx.gui import qt

from tomwer.core.process.control.datalistener.rpcserver import (
    BlissAcquisition,
    TangoAcquisitionStatus,
    _BaseDataListenerThread,
)
from tomwer.utils import docstring
from ._QThreadListenerBase import QThreadListenerBase

_logger = logging.getLogger(__name__)


class DataListenerQThread(_BaseDataListenerThread, QThreadListenerBase):
    """Implementation of the _BaseDataListenerThread with a QThread.

    As jsonrpc is using gevent we want to limit the 'crossover' of gevent and
    qt. So the server is launch in another process and bot processes are
    sharing list of acquisitions (added, ended...).
    """

    _WAIT_BTW_COLLECT = 0.5

    def __init__(self, host, port, acquisitions):
        qt.QThread.__init__(self)
        _BaseDataListenerThread.__init__(
            self, host=host, port=port, acquisitions=acquisitions
        )
        self._manager = Manager()

        self._lock = Lock()
        self._sequence_started = self._manager.list()
        self._sequence_ended = self._manager.list()
        self._scan_added = self._manager.list()
        self._stop = False

    @docstring(_BaseDataListenerThread.sequence_started)
    def sequence_started(self, acquisition: BlissAcquisition) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_started.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                )
            )

    def _collect_sequence_started(self):
        with self._lock:
            for scan in self._sequence_started:
                master_file, entry, proposal_file, sample_file = scan
                self.sigAcquisitionStarted.emit(
                    (master_file, entry, proposal_file, sample_file)
                )

            self._sequence_started[:] = []

    @docstring(_BaseDataListenerThread.scan_added)
    def scan_added(self, acquisition: BlissAcquisition, scan_number: int) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._scan_added.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    scan_number,
                )
            )

    def _collect_scan_added(self):
        with self._lock:
            for scan in self._scan_added:
                master_file, entry, proposal_file, sample_file, scan_number = scan
                self.sigScanAdded.emit(
                    (master_file, entry, proposal_file, sample_file, scan_number)
                )
            self._scan_added[:] = []

    @docstring(_BaseDataListenerThread.sequence_ended)
    def sequence_ended(self, acquisition: BlissAcquisition) -> None:
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_ended.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    not acquisition.has_error,
                )
            )

    def _collect_sequence_ended(self):
        with self._lock:
            for scan in self._sequence_ended:
                master_file, entry, proposal_file, sample_file, success = scan
                self.sigAcquisitionEnded.emit(
                    (master_file, entry, proposal_file, sample_file, success)
                )
            self._sequence_ended[:] = []

    @docstring(_BaseDataListenerThread.sequence_failed)
    def sequence_failed(self, acquisition: BlissAcquisition):
        assert isinstance(acquisition, BlissAcquisition)
        with self._lock:
            self._sequence_ended.append(
                (
                    acquisition.master_file,
                    acquisition.entry,
                    acquisition.proposal_file,
                    acquisition.sample_file,
                    False,
                )
            )

    def start(self):
        _BaseDataListenerThread.start(self)
        qt.QThread.start(self)

    def stop(self):
        _BaseDataListenerThread.stop(self)
        self._stop = True
        self.join()

    def _make_sure_server_still_running(self):
        if not self._server.is_alive():
            self.sigServerStop.emit()

    def run(self):
        while not self._stop:
            try:
                self._collect_sequence_started()
                self._collect_scan_added()
                self._collect_sequence_ended()
                self._make_sure_server_still_running()
            except Exception as e:
                _logger.error(e)
                break
            else:
                time.sleep(DataListenerQThread._WAIT_BTW_COLLECT)


_mock_acquisition_info = namedtuple(
    "_mock_acquisition_info", ["master_file", "entry", "scan_numbers", "waiting_time"]
)


class MockDataListenerQThread(DataListenerQThread):
    """
    Overwrite run() function and mock an acquisition for CI avoiding to have a
    tango server and tango install.
    """

    def __init__(self, host, port, acquisitions: list, mock_acquisitions: list):
        DataListenerQThread.__init__(
            self, host=host, port=port, acquisitions=acquisitions
        )
        assert isinstance(mock_acquisitions, (list, tuple))
        for acqui in mock_acquisitions:
            assert type(acqui) is _mock_acquisition_info
        self.mock_acquisitions = mock_acquisitions
        "waiting time between each scan / step of the acquisition"

    def get_device(self, device_name):
        return None

    def run(self) -> None:
        _logger.info("mock an acquisition using tango / tango")

        for mock_acqui in self.mock_acquisitions:
            if self._stop:
                return

            # first scan is the definition of the acquisition
            now = datetime.now()
            current_acquisition = BlissAcquisition(
                file_path=mock_acqui.master_file,
                entry_name=mock_acqui.entry,
                start_time=now.strftime("%H:%M:%S"),
                proposal_file=None,
                sample_file=None,
            )
            self.acquisitions.append(current_acquisition)
            self.sequence_started(current_acquisition)

            time.sleep(mock_acqui.waiting_time)

            # add scan numbers if any
            for scan_number in mock_acqui.scan_numbers:
                current_acquisition.add_scan_number(scan_number)
                current_acquisition.set_status(TangoAcquisitionStatus.ON_GOING)
                self.scan_added(
                    acquisition=current_acquisition, scan_number=scan_number
                )
                if self._stop:
                    return
                time.sleep(mock_acqui.waiting_time)

            # end acquisition
            now = datetime.now()
            current_acquisition.end(
                end_time=now.strftime("%H:%M:%S"), succeed=True, error=None
            )
            self._rpc_sequence_ended(
                saving_file=current_acquisition.sample_file,
                sequence_scan_number=current_acquisition.entry,
                success=True,
            )
            time.sleep(mock_acqui.waiting_time)

        # wait until stop the thread
        while not self._stop:
            time.sleep(0.1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to compute the center of rotation of a scan"""

import argparse
import functools
import logging
import signal
import sys
import silx
from silx.gui import qt

from tomwer.core.process.reconstruction.axis.axis import (
    AxisMode,
    AxisTask,
    NoAxisUrl,
)
from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.reconstruction.axis import AxisMainWindow
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.synctools.axis import QAxisRP

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class _AxisProcessGUI(AxisMainWindow):
    def __init__(self, scan, axis_params, parent=None, backend=None):
        axis_params.frame_width = scan.dim_1
        super().__init__(parent=parent, axis_params=axis_params, backend=backend)
        self.hideLockButton()
        self.hideApplyButton()
        self.setScan(scan=scan)

        # connect Signal / Slot
        callback = functools.partial(self.compute, scan)
        self.sigComputationRequested.connect(callback)
        self.setWindowIcon(icons.getQIcon("tomwer"))

    def compute(self, scan, wait=True):
        mess = " ".join(
            ("start axis calculation with", scan.axis_params.axis_url_1.url.path())
        )
        _logger.info(mess)
        process = AxisTask(
            inputs={
                "data": scan,
                "axis_params": self._axis_params,
                "wait": wait,
                "serialize_output_data": False,
            }
        )
        try:
            process.run()
        except NoAxisUrl:
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "Unable to find url to compute the axis, please select them "
                "from the `axis input` tab"
            )
            msg.setText(text)
            msg.exec()
            return None
        else:
            position = scan.axis_params.relative_cor_value
            assert isinstance(position, (float, type(None)))
            self.setPosition(relative_value=position)
            return position

    def _updatePosition(self, scan):
        self._widget.setPosition(relative_value=scan.axis_params.relative_cor_value)


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus "
        "provide the master file",
    )
    parser.add_argument(
        "--entry", help="For Nexus files: entry in the master file", default=None
    )
    parser.add_argument(
        "--use-sinogram",
        help="use the signoram from radio for computing COR. This only work"
        "with the =scan-path option",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--mode",
        default=None,
        help=f"Use a specific mode. Available modes are {[item.value for item in AxisMode]}",
    )
    parser.add_argument(
        "--full-image",
        action="store_true",
        default=False,
        help="Define the mode 'full' when display shifted images from manual" " mode",
    )
    parser.add_argument(
        "--use-opengl-plot",
        "--opengl-backend",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    mode = options.mode
    if mode is not None:
        mode = AxisMode.from_value(mode)

    try:
        scan = ScanFactory.create_scan_object(
            scan_path=options.scan_path, entry=options.entry
        )
    except ValueError:
        scan = ScanFactory.mock_scan()

    requires_reduced_dark_and_flat(scan=scan, logger_=_logger)
    # define the process_index is any tomwer_processes_existing
    if scan.axis_params is None:
        scan.axis_params = QAxisRP()

    if options.debug:
        _logger.setLevel(logging.DEBUG)

    axis_params = QAxisRP()

    window = _AxisProcessGUI(scan=scan, axis_params=axis_params)
    window.setWindowTitle("axis")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    if mode is not None:
        window.setMode(mode)
    if options.full_image:
        window.manual_uses_full_image(True)

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.exec()


def getinputinfo():
    return "tomwer axis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



"""Define nabu Axis"""

from enum import Enum as _Enum


class NabuPlane(_Enum):
    YZ = "YZ"
    XZ = "XZ"
    XY = "XY"

    @classmethod
    def from_value(cls, value):
        if value == 0:
            return NabuPlane.XY
        elif value == 1:
            return NabuPlane.XZ
        elif value == 2:
            return NabuPlane.YZ
        return NabuPlane(value=value)



from __future__ import annotations

from orangewidget import gui, settings, widget
from orangewidget.widget import Input, Output
from silx.gui import qt

from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.utils.deprecation import deprecated_warning


class NotifierWidgetOW(widget.OWBaseWidget):
    """
    simple widget which pop up and closes when recive a new object
    """

    name = "notifier"
    id = "orangecontrib.tomwer.widgets.control.NotifierOW"
    description = "Simple widget which pop up for 2 second when recives a new object"
    icon = "icons/notification.png"
    priority = 145
    keywords = ["control", "notifier", "notification"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = False

    _muted = settings.Setting(False)

    class Inputs:
        tomo_obj = Input(name="tomo_obj", type=TomwerObject, multiple=True)

    class Outputs:
        tomo_obj = Output(name="tomo_obj", type=TomwerObject)

    def __init__(self, parent=None):
        super().__init__(parent)
        deprecated_warning(
            type_="class",
            name="orangecontrib.tomwer.widgets.control.NotifierOW",
            replacement="orangecontrib.ewoksnotify.ToneNotifierOW",
            reason="moved to ewoksnotify (pip install ewoksnotify[full] to get the orange widget)",
            since_version="1.4",
        )
        self.pop_up = None
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._soundButton = qt.QPushButton(parent=self)
        self._soundButton.setMinimumSize(150, 100)
        self._soundButton.setCheckable(True)
        layout.addWidget(self._soundButton)
        self._updateButtonIcon()

        # connect signal / slot
        self._soundButton.toggled.connect(self._switchMute)

    def _switchMute(self):
        self._muted = not self._muted

        self._updateButtonIcon()

    @Inputs.tomo_obj
    def process(self, tomo_obj, *args, **kwargs):
        self.notify(tomo_obj)
        self.Outputs.tomo_obj.send(tomo_obj)

    def _updateButtonIcon(self):
        style = qt.QApplication.style()
        if self._muted:
            icon = style.standardIcon(qt.QStyle.SP_MediaVolumeMuted)
        else:
            icon = style.standardIcon(qt.QStyle.SP_MediaVolume)
        self._soundButton.setIcon(icon)

    def notify(self, tomo_obj):
        if self.pop_up is not None:
            self.pop_up.close()

        if not self._muted:
            # emit sound when requested
            try:
                qt.QApplication.beep()
            except Exception:
                pass

        self.pop_up = NotificationMessage()
        text = f"Object {tomo_obj} received."
        self.pop_up.setText(text)
        self.pop_up.show()


class NotificationMessage(qt.QMessageBox):
    EXPOSITION_TIME = 3000  # in ms

    def __init__(self) -> None:
        super().__init__()
        self.setModal(False)
        self.setIcon(qt.QMessageBox.Information)
        self.addButton(
            f"Ok - will close automatically after {self.EXPOSITION_TIME / 1000}s",
            qt.QMessageBox.YesRole,
        )

    def show(self):
        super().show()
        timer = qt.QTimer(self)
        timer.singleShot(
            self.EXPOSITION_TIME,
            self.close,
        )



from __future__ import annotations

import logging

from orangewidget import gui
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState
from silx.gui import qt

import tomwer.core.process.conditions.filters
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from tomwer.core.process.conditions import filters
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.conditions.filter import FileNameFilterWidget
from tomwer.utils import docstring

logger = logging.getLogger(__name__)


class NameFilterOW(SuperviseOW):
    name = "scan filter"
    id = "orange.widgets.tomwer.filterow"
    description = (
        "Simple widget which filter some data directory if the name"
        "doesn't match with the pattern defined."
    )
    icon = "icons/namefilter.svg"
    priority = 106
    keywords = ["tomography", "selection", "tomwer", "folder", "filter"]

    want_main_area = True
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.conditions.filters.FileNameFilterTask

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, multiple=True)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)

        self.widget = FileNameFilterWidget(parent=self)
        self.widget.setContentsMargins(0, 0, 0, 0)

        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        layout.addWidget(spacer)

    @Inputs.data
    def applyfilter(self, scan, *args, **kwargs):
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomwerScanBase}")

        process = filters.FileNameFilterTask(
            inputs={
                "data": scan,
                "pattern": self.getPattern(),
                "filter_type": self.getActiveFilter(),
                "invert_result": self.invertFilterAction(),
                "serialize_output_data": False,  # avoid spending time on scan serialization / deserialization when use orange
            }
        )
        process.run()
        out = process.outputs.data
        if out is not None:
            self.set_dataset_state(dataset=scan, state=DatasetState.SUCCEED)
            logger.processSucceed(f"{scan} pass through filter")
            self._signalScanReady(scan)
        else:
            self.set_dataset_state(dataset=scan, state=DatasetState.FAILED)
            logger.processFailed(f"{scan} NOT pass through filter")

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.applyfilter(dataset)

    def _signalScanReady(self, scan):
        self.Outputs.data.send(scan)

    def getPattern(self):
        return self.widget.getPattern()

    def setPattern(self, pattern):
        self.widget.setPattern(pattern)

    def getActiveFilter(self):
        return self.widget.getActiveFilter()

    def setActiveFilter(self, filter):
        self.widget.setActiveFilter(filter)

    def invertFilterAction(self):
        return self.widget.invertFilterAction()



from __future__ import annotations

from silx.gui import qt
from tomoscan.identifier import VolumeIdentifier
from tomwer.gui.visualization.reconstructionparameters import (
    ReconstructionParameters as _ReconstructionParameters,
)


class ReconstructionParameters(_ReconstructionParameters):
    """
    Add the scan identifier to the list of metadata displayed.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self._volumeIdentifierQLE = qt.QLineEdit("", self)
        self._volumeIdentifierQLE.setReadOnly(True)

        self.layout().insertRow(0, "volume id", self._volumeIdentifierQLE)

    def setVolumeIdentifier(self, volume_identifier: VolumeIdentifier):
        # TODO: connect the volume to the scan and display scan information as well
        # see https://gitlab.esrf.fr/tomotools/tomwer/-/issues/1479
        self._volumeIdentifierQLE.setText(volume_identifier.to_str())
        self._volumeIdentifierQLE.setToolTip(volume_identifier.to_str())



from __future__ import annotations

import signal

from silx.gui import qt
from tomwer.core.utils.process import send_signal_to_local_rpc_servers


class PortOccupyDialog(qt.QDialog):
    """Widget for the user to send a sig kill / sig term to the port listening to bliss-tomo TomoSync json-rpc code"""

    def __init__(self, parent, port, host):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._retry = False
        self.port = port
        self.host = host

        mess = (
            f"port ({port}) of {host} already in use. \n Maybe an other "
            "instance of `datalistener` is running in this session or "
            "another tomwer session. \n As this widget is connecting with "
            "bliss we enforce it to be unique."
        )
        self.layout().addWidget(qt.QLabel(mess, self))
        self.setWindowTitle("Unable to launch two listener in parallel")

        types = qt.QDialogButtonBox.Cancel | qt.QDialogButtonBox.Retry
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._sendSIGTERMMsb = qt.QPushButton("send SIGTERM", self)
        self._sendSIGTERMMsb.setToolTip(
            "Try to send SIGTERM signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGTERMMsb, qt.QDialogButtonBox.ActionRole)
        self._sendSIGKILLMsb = qt.QPushButton("send SIGKILL", self)
        self._sendSIGKILLMsb.setToolTip(
            "Try to send SIGKILL signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGKILLMsb, qt.QDialogButtonBox.ActionRole)

        # set up
        # for now we don't want to show "send signal" feature
        self._sendSIGTERMMsb.hide()
        self._sendSIGKILLMsb.hide()

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Cancel).released.connect(self.reject)
        self._buttons.button(qt.QDialogButtonBox.Retry).released.connect(
            self._retry_connect
        )
        self._sendSIGTERMMsb.released.connect(self._emitSigterm)
        self._sendSIGKILLMsb.released.connect(self._emitSigkill)

    @property
    def retry_connection(self):
        return self._retry

    def _emitSigterm(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGTERM, port=self.port)

    def _emitSigkill(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGKILL, port=self.port)

    def _retry_connect(self, *args, **kargs):
        self._retry = True
        self.accept()



import numpy
import pytest

from tomwer.core.utils.image import ImageScaleMethod, scale_img2_to_img1, shift_img


@pytest.mark.parametrize(
    "use_scipy, dx, dy, cval", ((True, 1.0, -2.3, 0), (False, 0.0, 3.6, 2))
)
def test_shift_image(use_scipy, dx, dy, cval):
    data = numpy.linspace(0, 12, 100 * 100).reshape(100, 100)
    assert isinstance(
        shift_img(data=data, dx=dx, dy=dy, cval=cval, use_scipy=use_scipy),
        numpy.ndarray,
    )


def test_scale_img2_to_img1():
    """
    test scale_img2_to_img1 function
    """
    img_1 = numpy.linspace(0, 10, 100 * 100, dtype=numpy.float16).reshape(100, 100)
    img_2 = numpy.linspace(20, 100, 100 * 100, dtype=numpy.float16).reshape(100, 100)

    rescale_img = scale_img2_to_img1(img_2, img_1, method=ImageScaleMethod.MEAN)
    numpy.testing.assert_array_almost_equal(
        img_2,
        rescale_img,
        decimal=1,
    )



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.control.datadiscovery import DataDiscoveryWidget
from tomwer.core.scan.scantype import ScanType


def test_data_discovery_widget(qtapp):  # noqa F811
    widget = DataDiscoveryWidget()
    pass
    widget.show()
    configuration = widget.getConfiguration()

    assert isinstance(configuration, dict)
    assert configuration == {
        "start_folder": "",
        "file_filter": None,
        "scan_type_searched": ScanType.NX_TOMO.value,
    }

    new_config = {
        "start_folder": "/my/folder",
        "file_filter": "*.nx",
        "scan_type_searched": ScanType.BLISS.value,
    }

    widget.setConfiguration(new_config)
    assert widget.getConfiguration() == new_config
    widget = None



from __future__ import annotations

from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.reconstruction.nabu.nabuconfig.nabuconfig import (
    NabuConfiguration,
)
from nabu.pipeline.config import _extract_nabuconfig_keyvals
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)


def test_NabuConfiguration(
    qtapp,  # noqa F811
):
    """Test that the 'NabuConfiguration' produce only keys understandable by nabu"""
    widget = NabuConfiguration(parent=None)

    tomwer_config = widget.getConfiguration()
    nabu_config = _extract_nabuconfig_keyvals(nabu_fullfield_default_config)

    tomwer_section_keys = set(tomwer_config.keys())
    nabu_section_keys = set(nabu_config.keys())
    tomwer_section_keys.remove("tomwer_slices")
    assert tomwer_section_keys.issubset(
        nabu_section_keys
    ), f"tomwer provide sections undefined by nabu: {tomwer_section_keys - nabu_section_keys}"

    for section_key in nabu_config.keys():
        if section_key in tomwer_config:
            nabu_sub_section_keys = set(nabu_config[section_key].keys())
            tomwer_sub_section_keys = set(tomwer_config[section_key].keys())
            if section_key == "phase":
                # this key is pop by tomwer before calling nabu. It is used internally only
                tomwer_sub_section_keys.remove("beam_shape")
            elif section_key == "reconstruction":
                tomwer_sub_section_keys.remove("slice_plane")
            elif section_key == "output":
                tomwer_sub_section_keys.remove("output_dir_mode")
            assert tomwer_sub_section_keys.issubset(
                nabu_sub_section_keys
            ), f"tomwer provides keys undefined by nabu: {tomwer_sub_section_keys - nabu_sub_section_keys}"



"""
(Ewoks)Orange Widgets for tomography raw and processed data visualization
"""

NAME = "visualization"

ID = "orangecontrib.tomwer.widgets.visualization"

DESCRIPTION = "widgets for data visualization"

LONG_DESCRIPTION = (
    "Contains widgets for visualization of the acquisitions," " reconstruction..."
)

ICON = "../../widgets/icons/tomwer_visu.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 3



# coding: utf-8
from __future__ import annotations


import functools
import logging

from silx.gui import qt
from silx.gui.dialog.DataFileDialog import DataFileDialog
from enum import Enum as _Enum

from nxtomomill.models.utils import convert_str_to_tuple
from tomwer.core.utils.char import DEGREE_CHAR
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSliceMode
from tomwer.core.process.reconstruction.nabu.utils import (
    _NabuFBPFilterType,
    _NabuPaddingType,
    _NabuReconstructionMethods,
    _NabuStages,
)
from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.gui.utils.inputwidget import SelectionLineEdit
from tomwer.gui.utils.scrollarea import (
    QComboBoxIgnoreWheel,
    QDoubleSpinBoxIgnoreWheel,
    QSpinBoxIgnoreWheel,
)
from tomwer.gui.utils.RangeWidget import RangeWidget
from tomwer.gui import icons
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class SliceGroupBox(qt.QGroupBox):
    """GroupBox to define slice to be reconstructed"""

    sigSlicesChanged = qt.Signal()
    """Signal emitted when the selected slices change"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "slices", parent)
        self.setCheckable(True)
        self.setLayout(qt.QHBoxLayout())
        # mode
        self._modeCB = QComboBoxIgnoreWheel(parent=self)
        for mode in NabuSliceMode:
            self._modeCB.addItem(mode.value)
        self.layout().addWidget(self._modeCB)
        self._modeCB.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)

        # slice line edit
        self._sliceQLE = SelectionLineEdit(parent=self, allow_negative_indices=True)
        self.layout().addWidget(self._sliceQLE)

        # warning icon
        warningIcon = icons.getQIcon("warning")
        self._noSliceWarning = qt.QLabel("")
        self._noSliceWarning.setPixmap(warningIcon.pixmap(20, state=qt.QIcon.On))
        self._noSliceWarning.setToolTip(
            "no slice defined, No slice reconstruction will be done"
        )
        self.layout().addWidget(self._noSliceWarning)

        # spacer
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # set up
        self.setChecked(True)
        self.setMode(NabuSliceMode.MIDDLE)
        self._sliceQLE.setVisible(False)
        self._noSliceWarning.setVisible(False)

        # connect signal / slot
        self._modeCB.currentIndexChanged.connect(self._updateSliceQLEVisibility)
        self._modeCB.currentIndexChanged.connect(self.sigSlicesChanged)
        self._sliceQLE.editingFinished.connect(self.sigSlicesChanged)
        self._sliceQLE.textChanged.connect(self._updateNoSliceWarningVisibility)
        self.toggled.connect(self.sigSlicesChanged)

    def _updateSliceQLEVisibility(self, *args, **kwargs):
        self._sliceQLE.setVisible(self.getMode() == NabuSliceMode.OTHER)
        self._updateNoSliceWarningVisibility()

    def _updateNoSliceWarningVisibility(self):
        def no_other_slice_defined() -> bool:
            slices_defined = self._sliceQLE.text().replace(" ", "")
            slices_defined = self._sliceQLE.text().replace(",", "")
            slices_defined = self._sliceQLE.text().replace(";", "")
            return slices_defined == ""

        self._noSliceWarning.setVisible(
            self.getMode() is NabuSliceMode.OTHER and no_other_slice_defined()
        )

    def setMode(self, mode: NabuSliceMode):
        mode = NabuSliceMode(mode)
        item_index = self._modeCB.findText(mode.value)
        self._modeCB.setCurrentIndex(item_index)
        old = self.blockSignals(True)
        self._updateSliceQLEVisibility()
        self.blockSignals(old)

    def getMode(self) -> NabuSliceMode:
        mode = NabuSliceMode(self._modeCB.currentText())
        return mode

    def _getSliceSelected(self):
        if self.getMode() is NabuSliceMode.MIDDLE:
            return NabuSliceMode.MIDDLE.value
        else:
            return self._sliceQLE.text()

    def getSlices(self):
        """Slice selected"""
        if self.isChecked():
            return self._getSliceSelected()
        else:
            return None

    def setSlices(self, slices):
        if slices is None:
            self.setChecked(False)
        else:
            self.setChecked(True)
            if slices != NabuSliceMode.MIDDLE.value:
                self._sliceQLE.setText(slices)
                self.setMode(NabuSliceMode.OTHER)
            else:
                self.setMode(NabuSliceMode.MIDDLE)


class TranslationMvtFileWidget(qt.QWidget):
    """Widget used to define a .cvs or a DataUrl"""

    class Mode(_Enum):
        HDF5 = "hdf5"
        TEXT = "text"

    fileChanged = qt.Signal()
    """Signal emitted when the file change"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())

        self._checkBox = qt.QCheckBox(self)
        self.layout().addWidget(self._checkBox)

        self._translationFileQLE = qt.QLineEdit("", self)
        self._translationFileQLE.setReadOnly(True)
        self.layout().addWidget(self._translationFileQLE)

        self._grpBox = qt.QGroupBox("file type", self)
        self._grpBox.setLayout(qt.QVBoxLayout())
        self._grpBox.layout().setContentsMargins(0, 0, 0, 0)
        self._grpBox.layout().setSpacing(0)

        self._hdf5FileRB = qt.QRadioButton(
            TranslationMvtFileWidget.Mode.HDF5.value, self
        )
        self._grpBox.layout().addWidget(self._hdf5FileRB)
        self._textFileRB = qt.QRadioButton(
            TranslationMvtFileWidget.Mode.TEXT.value, self
        )
        self._grpBox.layout().addWidget(self._textFileRB)
        self.layout().addWidget(self._grpBox)

        self._selectButton = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectButton)

        # set up
        self._grpBox.setEnabled(False)
        self._translationFileQLE.setEnabled(False)
        self._selectButton.setEnabled(False)
        self._textFileRB.setChecked(True)

        # connect signal / slot
        self._selectButton.released.connect(self._selectCallback)
        self._checkBox.toggled.connect(self._toggleSelection)

    def _selectCallback(self, *args, **kwargs):
        if self.getSelectionMode() is self.Mode.HDF5:
            file_or_url = self._selectHDF5()
        elif self.getSelectionMode() is self.Mode.TEXT:
            file_or_url = self._selectTextFile()
        else:
            raise ValueError("")
        if file_or_url is not None:
            self.setFile(file_or_url)

    def _selectHDF5(self):
        dialog = DataFileDialog()
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingDataset)

        if not dialog.exec():
            dialog.close()
            return
        else:
            return dialog.selectedUrl()

    def _selectTextFile(self):  # pragma: no cover
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFile)

        if not dialog.exec():
            dialog.close()
            return
        if len(dialog.selectedFiles()) > 0:
            return dialog.selectedFiles()[0]

    def isChecked(self):
        return self._checkBox.isChecked()

    def setChecked(self, checked):
        self._checkBox.setChecked(checked)

    def getSelectionMode(self):
        if self._hdf5FileRB.isChecked():
            return self.Mode.HDF5
        else:
            return self.Mode.TEXT

    def setFile(self, file_):
        if file_ in (None, ""):
            self.setChecked(False)
        else:
            self.setChecked(True)
            self._translationFileQLE.setText(file_)
            self.fileChanged.emit()

    def getFile(self):
        if self.isChecked():
            return self._translationFileQLE.text()
        else:
            return None

    def _toggleSelection(self):
        self._translationFileQLE.setEnabled(self._checkBox.isChecked())
        self._grpBox.setEnabled(self._checkBox.isChecked())
        self._selectButton.setEnabled(self._checkBox.isChecked())


class AnglesFileWidget(qt.QWidget):
    """widget to retrieve the text file for angles to provide to nabu"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setLayout(qt.QHBoxLayout())
        self._qle = qt.QLineEdit("", self)
        self.layout().addWidget(self._qle)
        self._selectPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectPB)

        # connect signal / slot
        self._selectPB.released.connect(self._selectFile)

    def getFile(self) -> str:
        return self._qle.text().replace(" ", "")

    def setFile(self, file_path):
        self._qle.setText(file_path)

    def _selectFile(self) -> str:  # pragma: no cover
        dialog = qt.QFileDialog()
        dialog.setFileMode(qt.QFileDialog.ExistingFile)
        dialog.setNameFilters(
            [
                "Text file (*.txt)",
                "Any file (*)",
            ]
        )

        if not dialog.exec():
            dialog.close()
            return
        elif len(dialog.selectedFiles()) == 0:
            return
        else:
            file_path = dialog.selectedFiles()[0]
            self._qle.setText(file_path)


class _NabuReconstructionConfig(qt.QWidget, _NabuStageConfigBase):
    """
    Widget to define the configuration of nabu reconstruction processing and dataset configuration.
    At the gui side the "dataset" section as no real reason to be on its own. So we melt the two there.
    """

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent, stage=_NabuStages.PRE)
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.PRE)
        self.setLayout(qt.QGridLayout())

        self.__optimizationIteAlgo = "chambolle-pock"

        # slices to be reconstructed online
        self._slicesWidget = SliceGroupBox(parent=self)
        self.layout().addWidget(self._slicesWidget, 0, 0, 1, 2)
        # ignore listener of the mouse for the slice widget, can
        # come with an issue as we are on a scroll area
        self._slicesWidget.setFocusPolicy(qt.Qt.FocusPolicy.NoFocus)

        # axis
        self._axisLabel = qt.QLabel("reconstruction plane", self)
        self._axisLabel.setToolTip("Over which axis the slice must be picked")
        self.layout().addWidget(self._axisLabel)
        self._axisQCB = QComboBoxIgnoreWheel(self)
        axis_tooltips = {
            NabuPlane.YZ.value: "along axis x (aka axis 2 - expected to be slow)",
            NabuPlane.XZ.value: "along axis y (aka axis 1 - expected to be slow)",
            NabuPlane.XY.value: "along axis Z (aka axis 0 - fastest)",
        }
        for item, tooltip in axis_tooltips.items():
            self._axisQCB.addItem(item)
            self._axisQCB.setItemData(
                self._axisQCB.findText(item),
                tooltip,
                qt.Qt.ToolTipRole,
            )
        self.layout().addWidget(self._axisQCB, 1, 1, 1, 1)

        # method
        self._methodLabel = qt.QLabel("method", self)
        self.layout().addWidget(self._methodLabel, 2, 0, 1, 1)
        self._methodQCB = QComboBoxIgnoreWheel(parent=self)
        for method in _NabuReconstructionMethods:
            self._methodQCB.addItem(method.value)
        self.layout().addWidget(self._methodQCB, 2, 1, 1, 1)
        self.registerWidget(self._methodLabel, "required")
        self.registerWidget(self._methodQCB, "required")

        self._methodWarningLabel = qt.QLabel(
            f"\tWARNING: \n\t1.' {_NabuReconstructionMethods.MLEM.value}' can be EXTREMELY SLOW on transmission\n\tdata and may produce erroneous results.\n\t2.MLEM should only be used with XRFCT data."
        )
        self._methodWarningLabel.setStyleSheet("color: red")
        self._methodWarningLabel.setVisible(False)
        self.layout().addWidget(self._methodWarningLabel, 3, 0, 1, 2)

        # angle_offset
        self._labelOffsetLabel = qt.QLabel("angle offset (in degree)", self)
        self.layout().addWidget(self._labelOffsetLabel, 4, 0, 1, 1)
        self._angleOffsetQDSB = QDoubleSpinBoxIgnoreWheel(self)
        self._angleOffsetQDSB.setMaximum(-180)
        self._angleOffsetQDSB.setMaximum(180)
        self.layout().addWidget(self._angleOffsetQDSB, 4, 1, 1, 1)
        self.registerWidget(self._labelOffsetLabel, "advanced")
        self.registerWidget(self._angleOffsetQDSB, "advanced")

        # fbp filter type
        self._fbpFilterCB = qt.QCheckBox("fbp filter", self)
        self.layout().addWidget(self._fbpFilterCB, 5, 0, 1, 1)
        self._fbpFilterType = QComboBoxIgnoreWheel(self)
        for filter_type in _NabuFBPFilterType:
            self._fbpFilterType.addItem(filter_type.value)
        self.layout().addWidget(self._fbpFilterType, 5, 1, 1, 1)
        self.registerWidget(self._fbpFilterCB, "advanced")
        self.registerWidget(self._fbpFilterType, "advanced")

        # padding type
        self._paddingTypeLabel = qt.QLabel("padding type", self)
        self.layout().addWidget(self._paddingTypeLabel, 6, 0, 1, 1)
        self._paddingType = QComboBoxIgnoreWheel(self)
        for fbp_padding_type in _NabuPaddingType:
            self._paddingType.addItem(fbp_padding_type.value)
        self.layout().addWidget(self._paddingType, 6, 1, 1, 1)
        self.registerWidget(self._paddingTypeLabel, "optional")
        self.registerWidget(self._paddingType, "optional")

        # sub region
        self._subRegionSelector = _NabuReconstructionSubRegion(parent=self)
        self.layout().addWidget(self._subRegionSelector, 8, 0, 1, 2)

        # iterations
        self._iterationsLabel = qt.QLabel("iterations", self)
        self.layout().addWidget(self._iterationsLabel, 9, 0, 1, 1)
        self._iterationSB = qt.QSpinBox(parent=self)
        self.layout().addWidget(self._iterationSB, 9, 1, 1, 1)
        self._iterationSB.setMinimum(1)
        self._iterationSB.setMaximum(9999)
        # not supported for now so hidden
        self._iterationsLabel.hide()
        self._iterationSB.hide()

        # binning - subsampling
        self._binSubSamplingGB = _BinSubSampling(
            "binning and sub-sampling", parent=self
        )
        self.layout().addWidget(self._binSubSamplingGB, 10, 0, 1, 2)

        # optimization algorithm:
        # set has default value for now, because has only one at the moment

        # weight total variation
        self._tvLabel = qt.QLabel("total variation weight", self)
        self.layout().addWidget(self._tvLabel, 11, 0, 1, 1)
        self._totalVariationWeight = qt.QDoubleSpinBox(self)
        self._totalVariationWeight.setMinimum(0.0)
        self._totalVariationWeight.setMaximum(1.0)
        self._totalVariationWeight.setDecimals(4)
        self._totalVariationWeight.setSingleStep(0.002)
        self.layout().addWidget(self._totalVariationWeight, 10, 1, 1, 1)
        # not supported for now so hidden
        self._tvLabel.hide()
        self._totalVariationWeight.hide()

        # preconditioning filter
        self._preconditioningFilter = qt.QCheckBox("preconditioning_filter", self)
        self._preconditioningFilter.setToolTip(
            'Whether to enable "filter ' 'preconditioning" for iterative' " methods"
        )
        self.layout().addWidget(self._preconditioningFilter, 11, 0, 1, 2)
        # not supported for now so hidden
        self._preconditioningFilter.hide()

        # positivity constraint
        self._positivityConstraintCB = qt.QCheckBox("positivity constraint", self)
        self._positivityConstraintCB.setToolTip(
            "Whether to enforce a " "positivity constraint in the " "reconstruction."
        )
        self.layout().addWidget(self._positivityConstraintCB, 12, 0, 1, 2)
        # not supported for now so hidden
        self._positivityConstraintCB.hide()

        # clip_outer_circle option
        self._clipOuterCircleCB = qt.QCheckBox("clip outer circle", self)
        self._clipOuterCircleCB.setToolTip(
            "Whether to set to zero voxels falling outside of the reconstruction region"
        )
        self.layout().addWidget(self._clipOuterCircleCB, 13, 0, 1, 2)
        self.registerWidget(self._clipOuterCircleCB, "optional")

        # centered axis option
        self._centeredAxisCB = qt.QCheckBox("centered axis", self)
        self._centeredAxisCB.setToolTip("")
        self.layout().addWidget(self._centeredAxisCB, 14, 0, 1, 2)
        self.registerWidget(self._centeredAxisCB, "optional")

        # translation movement file
        self._transMvtFileLabel = qt.QLabel("translation movement file", self)
        self.layout().addWidget(self._transMvtFileLabel, 22, 0, 1, 1)
        self._transMvtFileWidget = TranslationMvtFileWidget(self)
        self.layout().addWidget(self._transMvtFileWidget, 22, 1, 1, 1)
        self.registerWidget(self._transMvtFileLabel, "advanced")
        self.registerWidget(self._transMvtFileWidget, "advanced")
        translation_movement_tooltip = """
        A file where each line describes the horizontal and vertical translations of the sample (or detector).
        The order is 'horizontal, vertical'
        It can be created from a numpy array saved with 'numpy.savetxt'"""
        self._transMvtFileLabel.setToolTip(translation_movement_tooltip)
        self._transMvtFileWidget.setToolTip(translation_movement_tooltip)

        # angle files (if the user want's to overwrite rotation angles)
        self._angleFileLabel = qt.QLabel("angles file", self)
        self.layout().addWidget(self._angleFileLabel, 23, 0, 1, 1)
        self._anglesFileWidget = AnglesFileWidget(self)
        self.layout().addWidget(self._anglesFileWidget, 23, 1, 1, 1)
        self.registerWidget(self._angleFileLabel, "advanced")
        self.registerWidget(self._anglesFileWidget, "advanced")

        # exclude projections: [A, B] (if user want's to skip projection from degree A to degree B (included))
        self._excludeProjectionsQCB = qt.QCheckBox("exclude proj angles", self)
        self.layout().addWidget(self._excludeProjectionsQCB, 24, 0, 1, 1)
        self._excludeProjectionsWidget = RangeWidget(self)
        self._excludeProjectionsWidget.setRange(0.0, 360.0)
        self._excludeProjectionsWidget.setSuffix(DEGREE_CHAR)
        self._excludeProjectionsWidget.setToolTip(
            "Projections with a rotation in [a, b] range will be ignored."
        )
        self.layout().addWidget(self._excludeProjectionsWidget, 24, 1, 1, 1)
        self.registerWidget(self._excludeProjectionsQCB, "advanced")
        self.registerWidget(self._excludeProjectionsWidget, "advanced")

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 200, 1, 1, 1)

        # set up
        self.setNabuPlane("XY")
        self._fbpFilterCB.setChecked(True)
        fbp_item = self._methodQCB.findText(_NabuReconstructionMethods.FBP.value)
        self._methodQCB.setCurrentIndex(fbp_item)
        self._angleOffsetQDSB.setValue(0)
        ramlak_item = self._fbpFilterType.findText(_NabuFBPFilterType.RAMLAK.value)
        self._fbpFilterType.setCurrentIndex(ramlak_item)
        padding_type_item = self._paddingType.findText(_NabuPaddingType.ZEROS.value)
        self._fbpFilterType.setCurrentIndex(padding_type_item)
        self._iterationSB.setValue(200)
        self._totalVariationWeight.setValue(1.0e-2)
        self._preconditioningFilter.setChecked(True)
        self._positivityConstraintCB.setChecked(True)
        self._clipOuterCircleCB.setChecked(False)
        self._centeredAxisCB.setChecked(False)
        self._excludeProjectionsWidget.setEnabled(False)

        # connect signal / slot
        self._axisQCB.currentIndexChanged.connect(self._slicesChanged)
        self._methodQCB.currentIndexChanged.connect(self._methodChanged)
        self._angleOffsetQDSB.editingFinished.connect(self._angleOffsetChanged)
        self._fbpFilterCB.toggled.connect(self._FBPFilterTypeChanged)
        self._fbpFilterType.currentIndexChanged.connect(self._FBPFilterTypeChanged)
        self._paddingType.currentTextChanged.connect(self._paddingTypeChanged)
        self._subRegionSelector.sigConfChanged.connect(self._signalConfChanged)
        self._iterationSB.valueChanged.connect(self._nbIterationChanged)
        self._totalVariationWeight.valueChanged.connect(self._weightTvChanged)
        self._preconditioningFilter.toggled.connect(self._preconditionningFilterChanged)
        self._positivityConstraintCB.toggled.connect(self._positivityConstraintChanged)
        self._transMvtFileWidget.fileChanged.connect(self._mvtFileChanged)
        self._slicesWidget.sigSlicesChanged.connect(self._slicesChanged)
        self._binSubSamplingGB.binningChanged.connect(self._binningChanged)
        self._anglesFileWidget._qle.textChanged.connect(self._anglesFilechanged)
        self._clipOuterCircleCB.toggled.connect(self._clipOuterCircleChanged)
        self._centeredAxisCB.toggled.connect(self._centeredAxisChanged)
        self._excludeProjectionsQCB.toggled.connect(self._excludeProjectionsChanged)
        self._excludeProjectionsQCB.toggled.connect(
            self._excludeProjectionsWidget.setEnabled
        )
        self._excludeProjectionsWidget.sigChanged.connect(
            self._excludeProjectionsChanged
        )

    def getSlices(self):
        return self._slicesWidget.getSlices()

    def setSlices(self, slices):
        self._slicesWidget.setSlices(slices=slices)

    def hideSlicesInterface(self):
        self._slicesWidget.hide()

    def setConfigurationLevel(self, level):
        _NabuStageConfigBase.setConfigurationLevel(self, level)
        self._subRegionSelector.setConfigurationLevel(level=level)

    def _slicesChanged(self, *args, **kwargs):
        self._signalConfChanged("tomwer_slices")

    def _methodChanged(self, index):  # *args, **kwargs):
        self._signalConfChanged("method")
        selected_method = self.getMethod()
        self._methodWarningLabel.setVisible(
            selected_method == _NabuReconstructionMethods.MLEM
        )

    def _angleOffsetChanged(self, *args, **kwargs):
        self._signalConfChanged("angle_offset")

    def _FBPFilterTypeChanged(self, *args, **kwargs):
        self._signalConfChanged("fbp_filter_type")

    def _paddingTypeChanged(self, *args, **kwargs):
        self._signalConfChanged("padding_type")

    def _nbIterationChanged(self, *args, **kwargs):
        self._signalConfChanged("iterations")

    def _weightTvChanged(self, *args, **kwargs):
        self._signalConfChanged("weight_tv")

    def _preconditionningFilterChanged(self, *args, **kwargs):
        self._signalConfChanged("preconditioning_filter")

    def _positivityConstraintChanged(self, *args, **kwargs):
        self._signalConfChanged("positivity_constraint")

    def _mvtFileChanged(self, *args, **kwargs):
        self._signalConfChanged("translation_movements_file")

    def _clipOuterCircleChanged(self, *args, **kwargs):
        self._signalConfChanged("clip_outer_circle")

    def _centeredAxisChanged(self, *args, **kwargs):
        self._signalConfChanged("centered_axis")

    def _binningChanged(self, *args, **kwargs):
        self._signalConfChanged("binning")
        self._signalConfChanged("binning_z")

    def _anglesFilechanged(self, *args, **kwargs):
        self._signalConfChanged("angles_file")

    def _excludeProjectionsChanged(self):
        self._signalConfChanged("exclude_projections")

    def setScan(self, scan):
        raise NotImplementedError()

    def getMethod(self) -> _NabuReconstructionMethods:
        return _NabuReconstructionMethods(self._methodQCB.currentText())

    def setMethod(self, method):
        method = _NabuReconstructionMethods(method)
        item_index = self._methodQCB.findText(method.value)
        self._methodQCB.setCurrentIndex(item_index)

    def getAngleOffset(self) -> float:
        return self._angleOffsetQDSB.value()

    def setAngleOffset(self, value: str | float):
        self._angleOffsetQDSB.setValue(float(value))

    def getFBPFilterType(self) -> _NabuFBPFilterType | None:
        if self._fbpFilterCB.isChecked():
            return _NabuFBPFilterType(self._fbpFilterType.currentText())
        else:
            return None

    def setFBPFilterType(self, filter_type):
        if type(filter_type) is str and filter_type.lower() == "none":
            filter_type = None
        if filter_type is None:
            self._fbpFilterCB.setChecked(False)
        else:
            self._fbpFilterCB.setChecked(True)
            filter_type = _NabuFBPFilterType(filter_type)
            filter_index = self._fbpFilterType.findText(filter_type.value)
            self._fbpFilterType.setCurrentIndex(filter_index)

    def getFBPPaddingType(self) -> _NabuPaddingType:
        return _NabuPaddingType(self._paddingType.currentText())

    def setFBPPaddingType(self, padding):
        padding = _NabuPaddingType(padding)
        padding_index = self._paddingType.findText(padding.value)
        self._paddingType.setCurrentIndex(padding_index)

    def getNIterations(self):
        return self._iterationSB.value()

    def setNIterations(self, n_iterations):
        self._iterationSB.setValue(n_iterations)

    def getTotalVariationWeight(self) -> float:
        return self._totalVariationWeight.value()

    def setTotalVariationWeight(self, weight: float):
        self._totalVariationWeight.setValue(float(weight))

    def isPreconditioningFilterEnable(self):
        return self._preconditioningFilter.isChecked()

    def setPreconditioningFilterEnable(self, enable: bool | int):
        self._preconditioningFilter.setChecked(bool(enable))

    def isPositivityConstraintEnable(self):
        return self._positivityConstraintCB.isChecked()

    def setPositivityConstraintEnable(self, enable: bool | int):
        self._positivityConstraintCB.setChecked(bool(enable))

    def getTranslationMvtFile(self):
        return self._transMvtFileWidget.getFile()

    def setTranslationMvtFile(self, file_):
        self._transMvtFileWidget.setFile(file_)

    def getHorizontalBinning(self):
        return self._binSubSamplingGB.getHorizontalBinning()

    def setHorizontalBinning(self, binning):
        return self._binSubSamplingGB.setHorizontalBinning(binning=binning)

    def getVerticalBinning(self):
        return self._binSubSamplingGB.getVerticalBinning()

    def setVerticalBinning(self, binning):
        return self._binSubSamplingGB.setVerticalBinning(binning=binning)

    def getProjSubsampling(self):
        return self._binSubSamplingGB.getProjSubsampling()

    def setProjSubsampling(self, subsampling):
        return self._binSubSamplingGB.setProjSubsampling(subsampling=subsampling)

    def getAnglesFile(self) -> str:
        return self._anglesFileWidget.getFile()

    def setAnglesFile(self, angles_file: str):
        return self._anglesFileWidget.setFile(angles_file)

    def getClipOuterCircle(self) -> bool:
        return self._clipOuterCircleCB.isChecked()

    def setClipOuterCircle(self, checked: bool) -> None:
        self._clipOuterCircleCB.setChecked(checked)

    def getCenteredAxis(self) -> bool:
        return self._centeredAxisCB.isChecked()

    def setCenteredAxis(self, checked: bool):
        self._centeredAxisCB.setChecked(checked)

    def getNabuPlane(self) -> NabuPlane:
        """return over which axis we expect to do the reconstruction"""
        return NabuPlane.from_value(self._axisQCB.currentText())

    def setNabuPlane(self, axis: str | NabuPlane):
        axis = NabuPlane.from_value(axis)
        self._axisQCB.setCurrentText(axis.value)

    def getExcludeProjections(self) -> tuple:
        if self._excludeProjectionsQCB.isChecked():
            return self._excludeProjectionsWidget.getRange()
        else:
            return tuple()

    def setExcludeProjections(self, projections: tuple):
        if not isinstance(projections, tuple):
            raise TypeError(
                f"projections should be an instance of tuple. Got {type(projections)}"
            )
        if len(projections) == 0:
            self._excludeProjectionsQCB.setChecked(False)
        else:
            self._excludeProjectionsQCB.setChecked(True)
            self._excludeProjectionsWidget.setRange(*projections)

    @docstring(_NabuStageConfigBase)
    def getConfiguration(self) -> dict:
        fbp_filter_type = self.getFBPFilterType()
        if fbp_filter_type is None:
            fbp_filter_type = "none"
        else:
            fbp_filter_type = fbp_filter_type.value
        config = {
            "method": self.getMethod().value,
            "slice_plane": self.getNabuPlane().value,
            "angles_file": self.getAnglesFile(),
            "axis_correction_file": "",  # not managed for now
            "angle_offset": self.getAngleOffset(),
            "fbp_filter_type": fbp_filter_type,
            "padding_type": self.getFBPPaddingType().value,
            "iterations": self.getNIterations(),
            "optim_algorithm": self.__optimizationIteAlgo,
            "weight_tv": self.getTotalVariationWeight(),
            "preconditioning_filter": int(self.isPreconditioningFilterEnable()),
            "positivity_constraint": int(self.isPositivityConstraintEnable()),
            "rotation_axis_position": "",
            "translation_movements_file": self.getTranslationMvtFile() or "",
            "clip_outer_circle": int(self.getClipOuterCircle()),
            "centered_axis": int(self.getCenteredAxis()),
        }
        config.update(self._subRegionSelector.getConfiguration())
        return config

    def getDatasetConfiguration(self) -> dict:
        exclude_projections = self.getExcludeProjections()
        if len(exclude_projections) == 0:
            exclude_projections = ""
        else:
            exclude_projections = f"angular_range={str(list(exclude_projections))}"
        config = self._binSubSamplingGB.getConfiguration()
        config["exclude_projections"] = exclude_projections
        return config

    def setDatasetConfiguration(self, config):
        exclude_proj = config.pop("exclude_projections", None)
        if exclude_proj is not None:
            if isinstance(exclude_proj, str):
                exclude_proj = exclude_proj.replace("angular_range=", "")
            exclude_proj = convert_str_to_tuple(exclude_proj)
            self.setExcludeProjections(exclude_proj)
        return self._binSubSamplingGB.setConfiguration(config)

    @docstring(_NabuStageConfigBase)
    def setConfiguration(self, config):
        if "method" in config:
            self.setMethod(config["method"])
        if "angles_file" in config:
            self.setAnglesFile(config["angles_file"])
        if "angle_offset" in config:
            self.setAngleOffset(value=config["angle_offset"])
        if "fbp_filter_type" in config:
            self.setFBPFilterType(config["fbp_filter_type"])
        if "padding_type" in config:
            self.setFBPPaddingType(config["padding_type"])
        if "iterations" in config:
            self.setNIterations(int(config["iterations"]))
        if "optim_algorithm" in config:
            self.__optimizationIteAlgo = config["optim_algorithm"]
        if "weight_tv" in config:
            self.setTotalVariationWeight(weight=config["weight_tv"])
        if "preconditioning_filter" in config:
            self.setPreconditioningFilterEnable(int(config["preconditioning_filter"]))
        if "positivity_constraint" in config:
            self.setPositivityConstraintEnable(int(config["positivity_constraint"]))
        if "translation_movements_file" in config:
            self.setTranslationMvtFile(config["translation_movements_file"])
        if "clip_outer_circle" in config:
            self.setClipOuterCircle(bool(config["clip_outer_circle"]))
        if "centered_axis" in config:
            self.setCenteredAxis(bool(config["centered_axis"]))
        if "slice_plane" in config:
            self.setNabuPlane(config["slice_plane"])
        self._subRegionSelector.setConfiguration(config=config)

    def _signalConfChanged(self, param):
        self.sigConfChanged.emit(param)


class _SubRegionEditor(qt.QObject):
    sigConfChanged = qt.Signal(str)
    """Signal emitted each type a parameter is edited"""

    def __init__(
        self, parent, layout, layout_row: int, name: str, min_param: str, max_param: str
    ):
        assert type(layout_row) is int
        qt.QObject.__init__(self)
        self._layout = layout
        self.__minParam = min_param
        self.__maxParam = max_param

        validator = qt.QIntValidator()
        validator.setBottom(0)

        self._subRegionLabel = qt.QLabel(name, parent)
        self.layout().addWidget(self._subRegionLabel, layout_row, 0, 1, 1)

        # min
        self._minCB = qt.QCheckBox("min", parent)
        self.layout().addWidget(self._minCB, layout_row, 1, 1, 1)
        self._minQLE = qt.QLineEdit("0", parent)
        self._minQLE.setValidator(validator)
        self.layout().addWidget(self._minQLE, layout_row, 2, 1, 1)

        # max
        self._maxCB = qt.QCheckBox("max", parent)
        self.layout().addWidget(self._maxCB, layout_row, 3, 1, 1)
        self._maxQLE = qt.QLineEdit("0", parent)
        self._maxQLE.setValidator(validator)
        self.layout().addWidget(self._maxQLE, layout_row, 4, 1, 1)

        # spacer for style
        self._spacer = qt.QWidget(parent)
        self._spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(self._spacer, layout_row, 5, 1, 1)

        # set up
        self._minCB.setChecked(False)
        self._minQLE.setEnabled(False)
        self._maxCB.setChecked(False)
        self._maxQLE.setEnabled(False)

        # connect signal / slot
        self._minCB.toggled.connect(self._minQLE.setEnabled)
        self._maxCB.toggled.connect(self._maxQLE.setEnabled)
        self._minCB.toggled.connect(
            functools.partial(self._signalConfChanged, self.__minParam)
        )
        self._maxCB.toggled.connect(
            functools.partial(self._signalConfChanged, self.__maxParam)
        )
        self._minQLE.editingFinished.connect(
            functools.partial(self._signalConfChanged, self.__minParam)
        )
        self._maxQLE.editingFinished.connect(
            functools.partial(self._signalConfChanged, self.__maxParam)
        )

    def layout(self):
        return self._layout

    def getSubRegionMin(self) -> int | None:
        """

        :return: None if region is unbounded else the value of the bound
        """
        if self._minCB.isChecked():
            return int(self._minQLE.text())
        else:
            return None

    def getSubRegionMax(self) -> int | None:
        """

        :return: None if region is unbounded else the value of the bound
        """
        if self._maxCB.isChecked():
            return int(self._maxQLE.text())
        else:
            return None

    def setSubRegionMin(self, sub_region_min: int | None) -> None:
        """

        :param min: if min is None or -1 wr will expand it to - infinity
        """
        if sub_region_min in (-1, 0):
            sub_region_min = None
        if sub_region_min is None:
            self._minCB.setChecked(False)
        else:
            self._minCB.setChecked(True)
            self._minQLE.setText(str(sub_region_min))

    def setSubRegionMax(self, sub_region_max) -> None:
        """

        :param max: if max is None or -1 wr will expand it to infinity
        """
        if type(sub_region_max) is str:
            sub_region_max = int(sub_region_max)
        if sub_region_max == -1:
            sub_region_max = None
        if sub_region_max is None:
            self._maxCB.setChecked(False)
        else:
            self._maxCB.setChecked(True)
            self._maxQLE.setText(str(sub_region_max))

    def getSubRegion(self) -> tuple:
        """

        :return: min, max
        """
        return self.getSubRegionMin(), self.getSubRegionMax()

    def setSubRegion(
        self,
        sub_region_min: int | None,
        sub_region_max: int | None,
    ):
        self.setSubRegionMin(sub_region_min)
        self.setSubRegionMax(sub_region_max)

    def _signalConfChanged(self, param):
        self.sigConfChanged.emit(param)

    def setVisible(self, visible):
        for widget in (
            self._subRegionLabel,
            self._minCB,
            self._minQLE,
            self._maxCB,
            self._maxQLE,
            self._spacer,
        ):
            widget.setVisible(visible)


class _NabuReconstructionSubRegion(qt.QGroupBox, _NabuStageConfigBase):
    """Widget to select a sub region to reconstruct"""

    sigConfChanged = qt.Signal(str)
    """Signal emitted each type a parameter is edited"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent, stage=_NabuStages.PROC)
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.PROC)
        self.setTitle("sub region")

        self.setLayout(qt.QGridLayout())

        self._xSubRegion = _SubRegionEditor(
            parent=self,
            layout=self.layout(),
            layout_row=0,
            name="x",
            min_param="start_x",
            max_param="end_x",
        )
        self.registerWidget(self._xSubRegion, "optional")

        self._ySubRegion = _SubRegionEditor(
            parent=self,
            layout=self.layout(),
            layout_row=1,
            name="y",
            min_param="start_y",
            max_param="end_y",
        )
        self.registerWidget(self._ySubRegion, "optional")

        self._zSubRegion = _SubRegionEditor(
            parent=self,
            layout=self.layout(),
            layout_row=2,
            name="z",
            min_param="start_z",
            max_param="end_z",
        )
        self.registerWidget(self._zSubRegion, "optional")

        # set up

        # connect signal / slot
        self._xSubRegion.sigConfChanged.connect(self._signalConfChanged)
        self._ySubRegion.sigConfChanged.connect(self._signalConfChanged)
        self._zSubRegion.sigConfChanged.connect(self._signalConfChanged)

    def getConfiguration(self) -> dict:
        return {
            "start_x": self._xSubRegion.getSubRegionMin() or 0,
            "end_x": self._xSubRegion.getSubRegionMax() or -1,
            "start_y": self._ySubRegion.getSubRegionMin() or 0,
            "end_y": self._ySubRegion.getSubRegionMax() or -1,
            "start_z": self._zSubRegion.getSubRegionMin() or 0,
            "end_z": self._zSubRegion.getSubRegionMax() or -1,
        }

    def setConfiguration(self, config):
        if "start_x" in config:
            self._xSubRegion.setSubRegionMin(int(config["start_x"]))
        if "end_x" in config:
            self._xSubRegion.setSubRegionMax(int(config["end_x"]))
        if "start_y" in config:
            self._ySubRegion.setSubRegionMin(int(config["start_y"]))
        if "end_y" in config:
            self._ySubRegion.setSubRegionMax(int(config["end_y"]))
        if "start_z" in config:
            self._zSubRegion.setSubRegionMin(int(config["start_z"]))
        if "end_z" in config:
            self._zSubRegion.setSubRegionMax(int(config["end_z"]))

    def _signalConfChanged(self, param):
        self.sigConfChanged.emit(param)


class _BinSubSampling(qt.QGroupBox):
    binningChanged = qt.Signal()
    """signal emitted when binning change"""

    def __init__(self, text, parent):
        qt.QGroupBox.__init__(self, text)
        self.setLayout(qt.QFormLayout())
        # binning
        self._binningSB = QSpinBoxIgnoreWheel(self)
        self._binningSB.setMinimum(1)
        self._binningSB.setMaximum(3)
        self.layout().addRow("binning", self._binningSB)
        self._binningSB.setToolTip(
            "Binning factor in the horizontal dimension when reading the data. \nThe final slices dimensions will be divided by this factor."
        )
        # z binning
        self._zBinningSB = QSpinBoxIgnoreWheel(self)
        self._zBinningSB.setMinimum(1)
        self._zBinningSB.setMaximum(100)
        self.layout().addRow("vertical binning", self._zBinningSB)
        self._zBinningSB.setToolTip(
            "Binning factor in the vertical dimension when reading the data. \nThis results in a lesser number of reconstructed slices."
        )
        # projection sub-sampling
        self._projSubsamplingSB = QSpinBoxIgnoreWheel(self)
        self._projSubsamplingSB.setMinimum(1)
        self._projSubsamplingSB.setMaximum(100)
        self.layout().addRow("projection sub-sampling", self._projSubsamplingSB)

        # connect signal / slot
        self._zBinningSB.valueChanged.connect(self._valueUpdated)
        self._binningSB.valueChanged.connect(self._valueUpdated)

    def getHorizontalBinning(self) -> int:
        return self._binningSB.value()

    def setHorizontalBinning(self, binning: int):
        return self._binningSB.setValue(int(binning))

    def getVerticalBinning(self) -> int:
        return self._zBinningSB.value()

    def setVerticalBinning(self, binning: int):
        return self._zBinningSB.setValue(int(binning))

    def getProjSubsampling(self) -> int:
        return self._projSubsamplingSB.value()

    def setProjSubsampling(self, subsampling: int):
        return self._projSubsamplingSB.setValue(int(subsampling))

    def getConfiguration(self) -> dict:
        return {
            "binning": self.getHorizontalBinning(),
            "binning_z": self.getVerticalBinning(),
            "projections_subsampling": self.getProjSubsampling(),
        }

    def setConfiguration(self, config: dict):
        if "binning" in config:
            self.setHorizontalBinning(config["binning"])
        if "binning_z" in config:
            self.setVerticalBinning(config["binning_z"])
        if "projections_subsampling" in config:
            self.setProjSubsampling(subsampling=config["projections_subsampling"])

    def _valueUpdated(self, *args, **kwargs):
        self.binningChanged.emit()



from __future__ import annotations

import os
import shutil
import tempfile
import pint
import numpy

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.stitching.metadataholder import StitchingMetadata
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui.stitching.tests.utils import create_scans_z_series
from tomwer.gui.stitching.config.tomoobjdetails import TomoObjectPositionInfos
from tomwer.gui.stitching.axisorderedlist import AxisOrderedTomoObjsModel

_ureg = pint.get_application_registry()


class TestTomoObjectPositionInfos(TestCaseQt):
    """
    Test edition of the tomo object position information
    """

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self._y_pixel_size = 0.002
        self._x_pixel_size = 0.001
        self.scans = create_scans_z_series(
            self._tmp_path,
            z_positions_m=(0.200, 0.205, 0.210),
            x_positions_m=(0.0, 0.0, None),
            sample_pixel_size=(self._y_pixel_size, self._x_pixel_size),
            raw_frame_width=100,
        )
        self.scan_0_metadata = QStitchingMetadata(tomo_obj=self.scans[0])
        numpy.testing.assert_almost_equal(
            self.scan_0_metadata.get_raw_position_m(axis=0), 0.20
        )

        assert self.scan_0_metadata.get_raw_position_m(axis=2) == 0.0
        self.scan_2_metadata = QStitchingMetadata(tomo_obj=self.scans[2])
        numpy.testing.assert_almost_equal(
            self.scan_2_metadata.get_raw_position_m(axis=0), 0.21
        )
        self._widget = TomoObjectPositionInfos()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        shutil.rmtree(self._tmp_path)

    def testEdition(self):
        # test QStitchingMetadata are correctly read
        self._widget.show()
        self._widget.setStitchingMetadata(self.scan_0_metadata)
        self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(0.20 / 0.002)
        assert self._widget._axis_2_pos.get_final_px_position() == 0
        self._widget.setStitchingMetadata(self.scan_2_metadata)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(0.21 / 0.002)
        assert self._widget._axis_2_pos.get_final_px_position() is None
        # test editing some parameters from the GUI. Make sure GUI is updated as the underlying metadata object
        self._widget._axis_0_pos._overrideMetricPositionCB.setChecked(True)
        self._widget._axis_0_pos.setRawMetricPosition(
            position_m=0.4, displayed_unit=_ureg.nanometer
        )
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(
            0.4 * 10e-6 / 0.002
        )
        self._widget._axis_0_pos._overridePixelPositionCB.setChecked(True)
        self._widget._axis_0_pos.setOverrridePxPosition(position_px=120)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(120)
        assert self.scan_2_metadata.get_abs_position_px(axis=0) == int(120)


class TestAxisOrderedTomoObjsModel(TestCaseQt):
    """
    test the behavior of the z ordered model used for stitching
    """

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self.scans_with_pos_and_pixel_size = create_scans_z_series(
            os.path.join(self._tmp_path, "case1"),
            z_positions_m=(0.200, 0.205, 0.210),
            sample_pixel_size=0.001,
            raw_frame_width=100,
        )
        self.scans_with_pixel_size = create_scans_z_series(
            os.path.join(self._tmp_path, "case2"),
            z_positions_m=(None, None, None),
            sample_pixel_size=0.001,
            raw_frame_width=100,
        )
        self.scans_without_metadata = create_scans_z_series(
            os.path.join(self._tmp_path, "case3"),
            z_positions_m=(None, None, None),
            sample_pixel_size=None,
            raw_frame_width=100,
        )
        self._widget = qt.QTableView()
        self._widget.setModel(AxisOrderedTomoObjsModel(parent=self._widget, axis=0))

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        shutil.rmtree(self._tmp_path)

    def testWithMetadata(self):
        """
        make sure z ordering works in the best case scenario
        """
        first_expected_scan = (
            self.scans_with_pos_and_pixel_size[2].get_identifier().short_description()
        )
        # try giving it ordered
        for scan in self.scans_with_pos_and_pixel_size[::-1]:
            self._widget.model().addTomoObj(scan)
        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == first_expected_scan
        )
        self._widget.model().clearTomoObjs()
        # try giving it ordered (but inverted)
        for scan in self.scans_with_pos_and_pixel_size:
            self._widget.model().addTomoObj(scan)
        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == first_expected_scan
        )

    def testWithPixelSizeOnly(self):
        """
        make sure z ordering works when we are only aware about pixel size
        """
        for scan in self.scans_with_pixel_size[::-1]:
            self._widget.model().addTomoObj(scan)

        # try adding some metadata to object
        scans = self._widget.model()._objs
        scans = self._widget.model()._objs
        scans[0].stitching_metadata = StitchingMetadata(tomo_obj=scans[0])
        scans[1].stitching_metadata = QStitchingMetadata(tomo_obj=scans[1])

        scans[0].stitching_metadata._pos_as_m = (0.1, 1, 2)
        scans[1].stitching_metadata._pos_as_m = (-0.03, 1, 2)
        scans[2].stitching_metadata._pos_as_m = (0.23, 1, 2)

        self._widget.model().reorder_objs()
        self._widget.model().layoutChanged.emit()

        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == scans[2].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(1, 1), qt.Qt.DisplayRole
            )
            == scans[0].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(2, 1), qt.Qt.DisplayRole
            )
            == scans[1].get_identifier().short_description()
        )

    def testWithoutMetadata(self):
        """
        make sure tomo obj are still displayed even if there is no metadata at all
        """
        for scan in self.scans_without_metadata[::-1]:
            self._widget.model().addTomoObj(scan)
        # try adding some metadata to object
        scans = self._widget.model()._objs
        scans[0].stitching_metadata = StitchingMetadata(tomo_obj=scans[0])
        scans[1].stitching_metadata = QStitchingMetadata(tomo_obj=scans[1])

        scans[0].stitching_metadata._pos_as_px = (12, None, 2)
        scans[1].stitching_metadata._pos_as_px = (-12, None, None)
        scans[2].stitching_metadata._pos_as_px = (9999, None, None)

        self._widget.model().reorder_objs()
        self._widget.model().layoutChanged.emit()

        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == scans[2].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(1, 1), qt.Qt.DisplayRole
            )
            == scans[0].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(2, 1), qt.Qt.DisplayRole
            )
            == scans[1].get_identifier().short_description()
        )
